from dataclasses import dataclass

import numpy as np

from allencell_ml_segmenter.curation.main_view import (
    CurationMainView,
    MERGING_MASK_LAYER_NAME,
    EXCLUDING_MASK_LAYER_NAME,
)
from allencell_ml_segmenter._tests.fakes.fake_viewer import FakeViewer
from allencell_ml_segmenter.main.i_viewer import IViewer
from allencell_ml_segmenter.core.info_dialog_box import InfoDialogBox
from allencell_ml_segmenter.core.dialog_box import DialogBox
from qtpy.QtWidgets import QDialog
from allencell_ml_segmenter.curation.curation_model import (
    CurationModel,
    CurationView,
    CurationRecord,
    ImageType,
)
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter.core.image_data_extractor import ImageData
import allencell_ml_segmenter
from allencell_ml_segmenter.main.main_model import MainModel

from pytestqt.qtbot import QtBot
from unittest.mock import Mock
from pathlib import Path
import pytest
from pytest import MonkeyPatch

IMG_DIR_PATH = (
    Path(allencell_ml_segmenter.__file__).parent
    / "_tests"
    / "test_files"
    / "img_folder"
)

IMG_DIR_FILES = [path for path in IMG_DIR_PATH.iterdir()]
FAKE_IMG_DATA = [
    ImageData(28, 28, 28, 4, np.zeros([28, 28, 28, 4]), path)
    for path in IMG_DIR_FILES
]


@dataclass
class TestEnvironment:
    viewer: FakeViewer
    model: CurationModel
    view: CurationMainView


@pytest.fixture
def test_environment_with_seg2() -> TestEnvironment:
    curation_model: CurationModel = CurationModel(
        FakeExperimentsModel(), MainModel()
    )

    curation_model.set_image_directory_paths(ImageType.RAW, IMG_DIR_FILES)
    curation_model.set_image_directory_paths(ImageType.SEG1, IMG_DIR_FILES)
    curation_model.set_image_directory_paths(ImageType.SEG2, IMG_DIR_FILES)
    curation_model.set_current_view(CurationView.MAIN_VIEW)
    curation_model.start_loading_images()

    viewer: IViewer = FakeViewer()
    main_view: CurationMainView = CurationMainView(curation_model, viewer)

    return TestEnvironment(
        viewer,
        curation_model,
        main_view,
    )


@pytest.fixture
def test_environment_first_images_ready(
    test_environment_with_seg2: TestEnvironment,
) -> TestEnvironment:
    env: TestEnvironment = test_environment_with_seg2
    env.model.set_curr_image_data(ImageType.RAW, FAKE_IMG_DATA[0])
    env.model.set_curr_image_data(ImageType.SEG1, FAKE_IMG_DATA[0])
    env.model.set_curr_image_data(ImageType.SEG2, FAKE_IMG_DATA[0])

    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[1])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[1])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[1])
    return env


@pytest.fixture
def test_environment_without_seg2() -> TestEnvironment:
    curation_model: CurationModel = CurationModel(
        FakeExperimentsModel(), MainModel()
    )

    curation_model.set_image_directory_paths(ImageType.RAW, IMG_DIR_FILES)
    curation_model.set_image_directory_paths(ImageType.SEG1, IMG_DIR_FILES)
    curation_model.set_current_view(CurationView.MAIN_VIEW)
    curation_model.start_loading_images()

    viewer: IViewer = FakeViewer()
    main_view: CurationMainView = CurationMainView(curation_model, viewer)

    return TestEnvironment(
        viewer,
        curation_model,
        main_view,
    )


# NOTE: not using qtbot methods based on https://pytest-qt.readthedocs.io/en/latest/tutorial.html#note-about-qtbot-methods
### UI State Tests ----------------------------------------------------------------------------------


def test_initial_state_with_seg2(
    qtbot: QtBot, test_environment_with_seg2: TestEnvironment
) -> None:
    env: TestEnvironment = test_environment_with_seg2
    # Arrange
    loading_finished_mock: Mock = Mock()
    env.model.image_loading_finished.connect(loading_finished_mock)

    # everything should be disabled until signals from model fire
    # Assert
    assert not env.view.save_csv_button.isEnabled()
    assert not env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert not env.view.merging_base_combo.isEnabled()
    assert env.view.merging_base_combo.currentText() == "seg1"
    assert not env.view.merging_delete_button.isEnabled()

    assert not env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()

    assert env.view.use_img_stacked_spinner.is_spinning()
    assert not env.view.yes_radio.isEnabled()
    assert not env.view.no_radio.isEnabled()

    assert not env.view.next_button.isEnabled()

    # Act (pretend to be service)
    env.model.set_curr_image_data(ImageType.RAW, FAKE_IMG_DATA[0])
    env.model.set_curr_image_data(ImageType.SEG1, FAKE_IMG_DATA[0])
    env.model.set_curr_image_data(ImageType.SEG2, FAKE_IMG_DATA[0])

    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[1])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[1])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[1])

    # Assert
    # this behavior is already tested in the model tests, so if this fails, then you haven't
    # set all the required image data
    loading_finished_mock.assert_called_once()

    # save csv will not be enabled until >= 4 images have been curated
    assert not env.view.save_csv_button.isEnabled()
    assert env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert env.view.merging_base_combo.isEnabled()
    assert not env.view.merging_delete_button.isEnabled()

    assert env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()

    assert not env.view.use_img_stacked_spinner.is_spinning()
    assert env.view.yes_radio.isEnabled()
    assert env.view.no_radio.isEnabled()

    assert env.view.next_button.isEnabled()

    # TODO: these will fail if we change naming convention of added images...
    # is there a better way to check?
    assert env.viewer.contains_layer(f"[raw] {IMG_DIR_FILES[0].name}")
    assert env.viewer.contains_layer(f"[seg1] {IMG_DIR_FILES[0].name}")
    assert env.viewer.contains_layer(f"[seg2] {IMG_DIR_FILES[0].name}")

    assert env.view.progress_bar.value() == 1
    assert env.view.progress_bar.maximum() == len(IMG_DIR_FILES)


def test_initial_state_no_seg2(
    qtbot: QtBot, test_environment_without_seg2: TestEnvironment
) -> None:
    env: TestEnvironment = test_environment_without_seg2
    # Arrange
    loading_finished_mock: Mock = Mock()
    env.model.image_loading_finished.connect(loading_finished_mock)

    # everything should be disabled until signals from model fire
    # Assert
    assert not env.view.save_csv_button.isEnabled()
    assert not env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert not env.view.merging_base_combo.isEnabled()
    assert env.view.merging_base_combo.currentText() == "seg1"
    assert not env.view.merging_delete_button.isEnabled()

    assert not env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()

    assert env.view.use_img_stacked_spinner.is_spinning()
    assert not env.view.yes_radio.isEnabled()
    assert not env.view.no_radio.isEnabled()

    assert not env.view.next_button.isEnabled()

    # Act (pretend to be service)
    env.model.set_curr_image_data(ImageType.RAW, FAKE_IMG_DATA[0])
    env.model.set_curr_image_data(ImageType.SEG1, FAKE_IMG_DATA[0])

    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[1])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[1])

    # Assert
    loading_finished_mock.assert_called_once()

    # save csv will not be enabled until >= 4 images have been curated
    assert not env.view.save_csv_button.isEnabled()
    assert not env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert not env.view.merging_base_combo.isEnabled()
    assert not env.view.merging_delete_button.isEnabled()

    assert env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()

    assert not env.view.use_img_stacked_spinner.is_spinning()
    assert env.view.yes_radio.isEnabled()
    assert env.view.no_radio.isEnabled()

    assert env.view.next_button.isEnabled()

    assert env.viewer.contains_layer(f"[raw] {IMG_DIR_FILES[0].name}")
    assert env.viewer.contains_layer(f"[seg1] {IMG_DIR_FILES[0].name}")
    assert not env.viewer.contains_layer(f"[seg2] {IMG_DIR_FILES[0].name}")

    assert env.view.progress_bar.value() == 1
    assert env.view.progress_bar.maximum() == len(IMG_DIR_FILES)


def test_next_image(
    qtbot: QtBot, test_environment_with_seg2: TestEnvironment
) -> None:
    # Arrange
    env: TestEnvironment = test_environment_with_seg2
    loading_finished_mock: Mock = Mock()
    env.model.image_loading_finished.connect(loading_finished_mock)
    # finish 'loading' the first set of images
    env.model.set_curr_image_data(ImageType.RAW, FAKE_IMG_DATA[0])
    env.model.set_curr_image_data(ImageType.SEG1, FAKE_IMG_DATA[0])
    env.model.set_curr_image_data(ImageType.SEG2, FAKE_IMG_DATA[0])

    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[1])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[1])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[1])

    # Act
    env.view.next_button.click()

    # Assert
    assert loading_finished_mock.call_count == 1
    assert env.view.progress_bar.value() == 2

    assert env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert env.view.merging_base_combo.isEnabled()
    assert not env.view.merging_delete_button.isEnabled()

    assert env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()

    assert env.view.yes_radio.isEnabled()
    assert env.view.no_radio.isEnabled()

    assert not env.view.next_button.isEnabled()

    assert env.viewer.contains_layer(f"[raw] {IMG_DIR_FILES[1].name}")
    assert env.viewer.contains_layer(f"[seg1] {IMG_DIR_FILES[1].name}")
    assert env.viewer.contains_layer(f"[seg2] {IMG_DIR_FILES[1].name}")

    # Act
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[2])

    # Assert
    assert loading_finished_mock.call_count == 2
    assert env.view.next_button.isEnabled()


def test_last_image(
    qtbot: QtBot,
    test_environment_with_seg2: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    # Arrange
    # standard way to deal with modal dialogs: https://pytest-qt.readthedocs.io/en/latest/note_dialogs.html
    monkeypatch.setattr(InfoDialogBox, "exec", lambda *args: 0)
    env: TestEnvironment = test_environment_with_seg2
    loading_finished_mock: Mock = Mock()
    env.model.image_loading_finished.connect(loading_finished_mock)
    # finish 'loading' the first set of images
    env.model.set_curr_image_data(ImageType.RAW, FAKE_IMG_DATA[0])
    env.model.set_curr_image_data(ImageType.SEG1, FAKE_IMG_DATA[0])
    env.model.set_curr_image_data(ImageType.SEG2, FAKE_IMG_DATA[0])

    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[1])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[1])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[1])

    # Act
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[2])
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[3])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[3])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[3])
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[4])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[4])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[4])
    env.view.next_button.click()

    # Assert
    # one finished signal should be emitted after 'start_loading_images', and one
    # should be emitted after clicking the next button (which we do 4 times)
    assert loading_finished_mock.call_count == 5
    assert env.view.progress_bar.value() == 5

    assert env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert env.view.merging_base_combo.isEnabled()
    assert not env.view.merging_delete_button.isEnabled()

    assert env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()

    assert env.view.yes_radio.isEnabled()
    assert env.view.no_radio.isEnabled()

    assert env.view.next_button.isEnabled()

    assert env.viewer.contains_layer(f"[raw] {IMG_DIR_FILES[-1].name}")
    assert env.viewer.contains_layer(f"[seg1] {IMG_DIR_FILES[-1].name}")
    assert env.viewer.contains_layer(f"[seg2] {IMG_DIR_FILES[-1].name}")

    # Act
    env.view.next_button.click()

    # reached end, so buttons should be disabled
    # Assert
    assert loading_finished_mock.call_count == len(IMG_DIR_FILES)
    assert env.view.progress_bar.value() == len(IMG_DIR_FILES)

    assert not env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert not env.view.merging_base_combo.isEnabled()
    assert not env.view.merging_delete_button.isEnabled()

    assert not env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()

    assert not env.view.yes_radio.isEnabled()
    assert not env.view.no_radio.isEnabled()

    assert not env.view.next_button.isEnabled()


def test_create_new_merging_mask(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
) -> None:
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    assert len(env.viewer.get_all_shapes()) == 0

    # Act
    env.view.merging_create_button.click()

    # Assert
    assert len(env.viewer.get_all_shapes()) == 1
    assert env.viewer.get_shapes(MERGING_MASK_LAYER_NAME) is not None
    assert env.view.merging_save_button.isEnabled()
    assert env.view.merging_delete_button.isEnabled()


def test_create_new_excluding_mask(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
) -> None:
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    assert len(env.viewer.get_all_shapes()) == 0

    # Act
    env.view.excluding_create_button.click()

    # Assert
    assert len(env.viewer.get_all_shapes()) == 1
    assert env.viewer.get_shapes(EXCLUDING_MASK_LAYER_NAME) is not None
    assert env.view.excluding_save_button.isEnabled()
    assert env.view.excluding_delete_button.isEnabled()


def test_save_csv(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
) -> None:
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    save_requested_slot: Mock = Mock()
    env.model.save_to_disk_requested.connect(save_requested_slot)
    # need to get to a point where enough images are marked to use that saving is allowed
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[2])
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[3])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[3])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[3])
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[4])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[4])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[4])

    # Act / Assert
    env.view.save_csv_button.click()
    assert not env.view.save_csv_button.isEnabled()
    save_requested_slot.assert_called_once()

    env.model.set_curation_record_saved_to_disk(
        True
    )  # should re-enable the button
    assert env.view.save_csv_button.isEnabled()


def test_save_csv_enabled_state(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
) -> None:
    """
    Test that save csv button is enabled only when at least 4 images have been selected for use.
    """
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Assert
    assert not env.view.save_csv_button.isEnabled()

    # Act
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[2])
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[3])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[3])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[3])

    # Assert
    # we are at the 3rd image, so button should still be disabled
    assert not env.view.save_csv_button.isEnabled()

    # Act
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[4])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[4])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[4])

    # Assert
    # by default, use_image should be selected, so save csv button should be enabled
    assert env.view.save_csv_button.isEnabled()

    # Act
    env.view.no_radio.click()

    # Assert
    assert not env.view.save_csv_button.isEnabled()

    # Act
    env.view.yes_radio.click()
    env.view.next_button.click()

    # Assert
    assert env.view.save_csv_button.isEnabled()

    # Act
    env.view.no_radio.click()

    # Assert
    # since we are toggling between 4 and 5 images selected for use, should still be enabled
    assert env.view.save_csv_button.isEnabled()


def test_radio_button_enabled_state(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
) -> None:
    """
    Test that radio buttons get disabled when it's only possible to select 4 images for use.
    """
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Assert
    assert env.view.yes_radio.isEnabled()
    assert env.view.no_radio.isEnabled()

    # Act
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[2])

    # Assert
    assert env.view.yes_radio.isEnabled()
    assert env.view.no_radio.isEnabled()

    # Act
    env.view.no_radio.click()
    env.view.next_button.click()

    # Assert
    # since there are 5 images total, and we've said we don't want to use one, we shouldn't be able to
    # select not to use any more of them
    assert not env.view.yes_radio.isEnabled()
    assert not env.view.no_radio.isEnabled()


def test_delete_merging_mask(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
):
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act
    env.view.merging_create_button.click()
    env.view.merging_delete_button.click()

    # Assert
    assert len(env.viewer.get_all_shapes()) == 0
    assert env.viewer.get_shapes(MERGING_MASK_LAYER_NAME) is None
    assert not env.view.merging_delete_button.isEnabled()


def test_delete_excluding_mask(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
):
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act
    env.view.excluding_create_button.click()
    env.view.excluding_delete_button.click()

    # Assert
    assert len(env.viewer.get_all_shapes()) == 0
    assert env.viewer.get_shapes(EXCLUDING_MASK_LAYER_NAME) is None
    assert not env.view.excluding_delete_button.isEnabled()


def test_use_image_radios(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
) -> None:
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act / Assert
    assert env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert env.view.merging_base_combo.isEnabled()
    assert not env.view.merging_delete_button.isEnabled()

    assert env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()

    env.view.no_radio.click()

    assert not env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert not env.view.merging_base_combo.isEnabled()
    assert not env.view.merging_delete_button.isEnabled()

    assert not env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()

    env.view.yes_radio.click()

    assert env.view.merging_create_button.isEnabled()
    assert not env.view.merging_save_button.isEnabled()
    assert env.view.merging_base_combo.isEnabled()
    assert not env.view.merging_delete_button.isEnabled()

    assert env.view.excluding_create_button.isEnabled()
    assert not env.view.excluding_delete_button.isEnabled()
    assert not env.view.excluding_save_button.isEnabled()


### Model State Tests ------------------------------------------------------------------------


def test_set_use_image(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
):
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act / Assert
    assert env.view.yes_radio.isChecked()
    assert env.model.get_use_image()

    env.view.no_radio.click()
    assert not env.model.get_use_image()

    env.view.yes_radio.click()
    assert env.model.get_use_image()


def test_set_base_image(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
):
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act / Assert
    assert (
        env.model.get_base_image() == env.view.merging_base_combo.currentText()
    )
    env.view.merging_base_combo.setCurrentIndex(1)
    assert (
        env.model.get_base_image() == env.view.merging_base_combo.currentText()
    )
    env.view.merging_base_combo.setCurrentIndex(0)
    assert (
        env.model.get_base_image() == env.view.merging_base_combo.currentText()
    )


def test_set_merging_mask(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
):
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act / Assert
    env.view.merging_create_button.click()
    assert env.model.get_merging_mask() is None
    env.view.merging_save_button.click()
    assert env.model.get_merging_mask() is not None
    env.view.merging_delete_button.click()
    assert env.model.get_merging_mask() is None


def test_set_excluding_mask(
    qtbot: QtBot, test_environment_first_images_ready: TestEnvironment
):
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act / Assert
    env.view.excluding_create_button.click()
    assert env.model.get_excluding_mask() is None
    env.view.excluding_save_button.click()
    assert env.model.get_excluding_mask() is not None
    env.view.excluding_delete_button.click()
    assert env.model.get_excluding_mask() is None


def test_curation_record_on_next(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    # Arrange
    # standard way to deal with modal dialogs: https://pytest-qt.readthedocs.io/en/latest/note_dialogs.html
    monkeypatch.setattr(InfoDialogBox, "exec", lambda *args: 0)
    env: TestEnvironment = test_environment_first_images_ready

    # Act
    env.view.yes_radio.click()
    env.view.merging_create_button.click()
    env.view.merging_save_button.click()
    env.view.merging_base_combo.setCurrentIndex(0)
    merging_base: str = env.view.merging_base_combo.currentText()
    env.view.next_button.click()
    env.model.set_next_image_data(ImageType.RAW, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG1, FAKE_IMG_DATA[2])
    env.model.set_next_image_data(ImageType.SEG2, FAKE_IMG_DATA[2])

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert record.to_use
    assert record.excluding_mask is None
    assert record.merging_mask is not None
    assert record.seg1 == IMG_DIR_FILES[0]
    assert record.seg2 == IMG_DIR_FILES[0]
    assert record.raw_file == IMG_DIR_FILES[0]
    assert record.base_image == merging_base

    # Act
    env.view.no_radio.click()
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[1]
    assert not record.to_use
    assert record.excluding_mask is None
    assert record.merging_mask is None
    assert record.seg1 == IMG_DIR_FILES[1]
    assert record.seg2 == IMG_DIR_FILES[1]
    assert record.raw_file == IMG_DIR_FILES[1]
    assert record.base_image == "seg1"  # should default to seg1

    # Act
    env.view.yes_radio.click()
    env.view.excluding_create_button.click()
    env.view.excluding_save_button.click()
    env.view.merging_base_combo.setCurrentIndex(1)
    merging_base: str = env.view.merging_base_combo.currentText()
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[2]
    assert record.to_use
    assert record.excluding_mask is not None
    assert record.merging_mask is None
    assert record.seg1 == IMG_DIR_FILES[2]
    assert record.seg2 == IMG_DIR_FILES[2]
    assert record.raw_file == IMG_DIR_FILES[2]
    assert record.base_image == merging_base


def test_merg_viewer_unsaved_model_saved_yes_to_prompt(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: merging mask in viewer is unsaved, and mask in viewer is different from previously saved mask. User
    clicks 'next' then responds 'yes' to the 'do you want to save the unsaved changes' prompt.

    Expectation: we expect that the mask in the viewer will be saved and persist in model state.
    """
    # Arrange
    monkeypatch.setattr(
        DialogBox, "exec", lambda *args: QDialog.DialogCode.Accepted
    )
    env: TestEnvironment = test_environment_first_images_ready
    unsaved_mask: np.ndarray = np.asarray([[4, 5], [6, 7]])

    # Act
    env.view.merging_create_button.click()
    env.view.merging_save_button.click()
    env.viewer.modify_shapes(MERGING_MASK_LAYER_NAME, unsaved_mask)
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert np.array_equal(record.merging_mask, unsaved_mask)


def test_merg_viewer_unsaved_model_saved_no_to_prompt(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: merging mask in viewer is unsaved, and mask in viewer is different from previously saved mask. User
    clicks 'next' then responds 'no' to the 'do you want to save the unsaved changes' prompt.

    Expectation: we expect that the mask in the viewer will not be saved and the previously saved mask will
    persist in model state.
    """
    # Arrange
    monkeypatch.setattr(
        DialogBox, "exec", lambda *args: QDialog.DialogCode.Rejected
    )
    env: TestEnvironment = test_environment_first_images_ready
    unsaved_mask: np.ndarray = np.asarray([[4, 5], [6, 7]])

    # Act
    env.view.merging_create_button.click()
    env.view.merging_save_button.click()
    env.viewer.modify_shapes(MERGING_MASK_LAYER_NAME, unsaved_mask)
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert not np.array_equal(record.merging_mask, unsaved_mask)


def test_merg_viewer_unsaved_model_unsaved_yes_to_prompt(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: mask in viewer is unsaved, and there is *no previously saved mask*. User
    clicks 'next' then responds 'yes' to the 'do you want to save the unsaved changes' prompt.

    Expectation: we expect that the mask in the viewer will be saved and persist in model state.
    """
    # Arrange
    monkeypatch.setattr(
        DialogBox, "exec", lambda *args: QDialog.DialogCode.Accepted
    )
    env: TestEnvironment = test_environment_first_images_ready
    unsaved_mask: np.ndarray = np.asarray([[4, 5], [6, 7]])

    # Act
    env.view.merging_create_button.click()
    env.viewer.modify_shapes(MERGING_MASK_LAYER_NAME, unsaved_mask)
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert np.array_equal(record.merging_mask, unsaved_mask)


def test_merg_viewer_empty_model_unsaved(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: there is an empty merging mask in the viewer and no mask saved yet

    Expectation: user should click next and have no dialog pop up / no mask get saved.
    """
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act
    env.view.merging_create_button.click()
    # make it so that merging mask has no shapes
    env.viewer.modify_shapes(MERGING_MASK_LAYER_NAME, np.asarray([]))
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert record.merging_mask is None


def test_merg_viewer_empty_model_saved_yes_to_prompt(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: there is an empty merging mask in the viewer and some non-empty mask saved

    Expectation: clicking next causes dialog to pop up; empty mask will get saved when we select 'yes'.
    """
    # Arrange
    monkeypatch.setattr(
        DialogBox, "exec", lambda *args: QDialog.DialogCode.Accepted
    )
    env: TestEnvironment = test_environment_first_images_ready

    # Act
    env.view.merging_create_button.click()
    env.view.merging_save_button.click()
    # make it so that merging mask has no shapes
    env.viewer.modify_shapes(MERGING_MASK_LAYER_NAME, np.array([]))
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert len(record.merging_mask) == 0


def test_merg_viewer_unsaved_not_to_use(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: there is an unsaved merging mask in the viewer, but the user has opted not to use this image

    Expectation: no dialog will pop up since we are not using the image anyways
    """
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act
    env.view.merging_create_button.click()
    env.view.no_radio.click()
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert record.merging_mask is None


def test_exclud_viewer_unsaved_model_saved_yes_to_prompt(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: excluding mask in viewer is unsaved, and mask in viewer is different from previously saved mask. User
    clicks 'next' then responds 'yes' to the 'do you want to save the unsaved changes' prompt.

    Expectation: we expect that the mask in the viewer will be saved and persist in model state.
    """
    # Arrange
    monkeypatch.setattr(
        DialogBox, "exec", lambda *args: QDialog.DialogCode.Accepted
    )
    env: TestEnvironment = test_environment_first_images_ready
    unsaved_mask: np.ndarray = np.asarray([[4, 5], [6, 7]])

    # Act
    env.view.excluding_create_button.click()
    env.view.excluding_save_button.click()
    env.viewer.modify_shapes(EXCLUDING_MASK_LAYER_NAME, unsaved_mask)
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert np.array_equal(record.excluding_mask, unsaved_mask)


def test_exclud_viewer_unsaved_model_saved_no_to_prompt(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: excluding mask in viewer is unsaved, and mask in viewer is different from previously saved mask. User
    clicks 'next' then responds 'no' to the 'do you want to save the unsaved changes' prompt.

    Expectation: we expect that the mask in the viewer will not be saved and the previously saved mask will
    persist in model state.
    """
    # Arrange
    monkeypatch.setattr(
        DialogBox, "exec", lambda *args: QDialog.DialogCode.Rejected
    )
    env: TestEnvironment = test_environment_first_images_ready
    unsaved_mask: np.ndarray = np.asarray([[4, 5], [6, 7]])

    # Act
    env.view.excluding_create_button.click()
    env.view.excluding_save_button.click()
    env.viewer.modify_shapes(EXCLUDING_MASK_LAYER_NAME, unsaved_mask)
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert not np.array_equal(record.excluding_mask, unsaved_mask)


def test_exclud_viewer_unsaved_model_unsaved_yes_to_prompt(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: excluding mask in viewer is unsaved, and there is *no previously saved mask*. User
    clicks 'next' then responds 'yes' to the 'do you want to save the unsaved changes' prompt.

    Expectation: we expect that the mask in the viewer will be saved and persist in model state.
    """
    # Arrange
    monkeypatch.setattr(
        DialogBox, "exec", lambda *args: QDialog.DialogCode.Accepted
    )
    env: TestEnvironment = test_environment_first_images_ready
    unsaved_mask: np.ndarray = np.asarray([[4, 5], [6, 7]])

    # Act
    env.view.excluding_create_button.click()
    env.viewer.modify_shapes(EXCLUDING_MASK_LAYER_NAME, unsaved_mask)
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert np.array_equal(record.excluding_mask, unsaved_mask)


def test_exclud_viewer_empty_model_unsaved(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: there is an empty excluding mask in the viewer and no mask saved yet

    Expectation: user should click next and have no dialog pop up / no mask get saved.
    """
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act
    env.view.excluding_create_button.click()
    # make it so that excluding mask has no shapes
    env.viewer.modify_shapes(EXCLUDING_MASK_LAYER_NAME, np.asarray([]))
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert record.excluding_mask is None


def test_exclud_viewer_empty_model_saved_yes_to_prompt(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: there is an empty excluding mask in the viewer and some non-empty mask saved

    Expectation: clicking next causes dialog to pop up; empty mask will get saved when we select 'yes'.
    """
    # Arrange
    monkeypatch.setattr(
        DialogBox, "exec", lambda *args: QDialog.DialogCode.Accepted
    )
    env: TestEnvironment = test_environment_first_images_ready

    # Act
    env.view.excluding_create_button.click()
    env.view.excluding_save_button.click()
    # make it so that excluding mask has no shapes
    env.viewer.modify_shapes(EXCLUDING_MASK_LAYER_NAME, np.array([]))
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert len(record.excluding_mask) == 0


def test_exclud_viewer_unsaved_not_to_use(
    qtbot: QtBot,
    test_environment_first_images_ready: TestEnvironment,
    monkeypatch: MonkeyPatch,
) -> None:
    """
    Scenario: there is an unsaved excluding mask in the viewer, but the user has opted not to use this image

    Expectation: no dialog will pop up since we are not using the image anyways
    """
    # Arrange
    env: TestEnvironment = test_environment_first_images_ready

    # Act
    env.view.excluding_create_button.click()
    env.view.no_radio.click()
    env.view.next_button.click()

    # Assert
    record: CurationRecord = env.model.get_curation_record()[0]
    assert record.excluding_mask is None



from allencell_ml_segmenter.core.task_executor import ITaskExecutor
from typing import Callable, Optional, Any
from napari.qt.threading import FunctionWorker, create_worker  # type: ignore


class NapariThreadTaskExecutor(ITaskExecutor):

    _instance = None

    def exec(
        self,
        task: Callable[[], Any],
        on_start: Optional[Callable[[], Any]] = None,
        on_finish: Optional[Callable[[], None]] = None,
        on_return: Optional[Callable[[Any], None]] = None,
        on_error: Optional[Callable[[Exception], None]] = None,
    ) -> None:
        worker: FunctionWorker = create_worker(task)
        if on_start is not None:
            worker.started.connect(on_start)
        if on_finish is not None:
            worker.finished.connect(on_finish)
        if on_return is not None:
            worker.returned.connect(on_return)
        if on_error is not None:
            worker.errored.connect(on_error)
        worker.start()

    @classmethod
    def global_instance(cls) -> ITaskExecutor:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance



import os

from qtpy.QtWidgets import QFileDialog


class DirectoryOrCSVFileDialog(QFileDialog):
    """
    A custom QFileDialog that allows the user to select either a directory or a CSV file.
    Used in relation to InputButton.
    """

    def __init__(self) -> None:
        super().__init__()
        self.setOption(QFileDialog.DontUseNativeDialog)
        self.setFileMode(QFileDialog.Directory)
        self.currentChanged.connect(self._selected)
        self.setNameFilter("Directories and CSV files (*.csv)")

    def _selected(self, name: str) -> None:
        """
        Called whenever the user selects a new option in the File Dialog menu.
        """
        if os.path.isdir(name):
            self.setFileMode(QFileDialog.Directory)
            self.setNameFilter(
                "Directories and CSV files (*.csv)"
            )  # without this, the filter is changed to just "Directories"
        else:
            self.setFileMode(QFileDialog.ExistingFile)



from pathlib import Path
from typing import Dict, Union

import pytest
from allencell_ml_segmenter._tests.fakes.fake_user_settings import (
    FakeUserSettings,
)
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ImageType,
)
from allencell_ml_segmenter.utils.cyto_overrides_manager import (
    CytoDLOverridesManager,
)
import allencell_ml_segmenter


@pytest.fixture
def experiments_model() -> ExperimentsModel:
    exp_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "main"
        / "experiments_home"
    )
    experiments_model = ExperimentsModel(
        FakeUserSettings(
            cyto_dl_home_path=Path(), user_experiments_path=exp_path
        )
    )
    return experiments_model


@pytest.fixture
def training_model(experiments_model: ExperimentsModel) -> TrainingModel:
    model: TrainingModel = TrainingModel(MainModel(), experiments_model)
    model.set_experiment_type("segmentation")
    model.set_spatial_dims(3)
    model.set_images_directory("/path/to/images")
    model.set_selected_channel(ImageType.RAW, 1)
    model.set_selected_channel(ImageType.SEG1, 2)
    model.set_selected_channel(ImageType.SEG2, 3)
    model.set_use_max_time(True)
    model.set_max_time(9992)
    model.set_patch_size([1, 2, 4])
    model.set_num_epochs(100)
    model.set_model_size("medium")
    return model


def test_get_training_overrides(
    experiments_model: ExperimentsModel, training_model: TrainingModel
):
    experiments_model.apply_experiment_name("one_ckpt_exp")
    cyto_overrides_manager: CytoDLOverridesManager = CytoDLOverridesManager(
        experiments_model, training_model
    )

    training_overrides: Dict[str, Union[str, int, float, bool, Dict]] = (
        cyto_overrides_manager.get_training_overrides()
    )

    assert (
        training_overrides["spatial_dims"] == training_model.get_spatial_dims()
    )

    # the experiments model fixture is set to the experiment @ _tests/experiments_home/one_ckpt_exp,
    # for which the last 'best' checkpoint is epoch_000.ckpt. Since we are currently at the first epoch checkpoint,
    # in order to run training_model.get_num_epochs() more epochs, we need to increment by 1
    assert (
        training_overrides["trainer.max_epochs"]
        == training_model.get_num_epochs() + 1
    )
    assert (
        training_overrides["trainer.max_time"]["minutes"]
        == training_model.get_max_time()
    )

    assert training_overrides["data.path"] == str(
        training_model.get_images_directory()
    )

    assert (
        training_overrides["data._aux.patch_shape"]
        == training_model.get_patch_size()
    )

    assert training_overrides[
        "input_channel"
    ] == training_model.get_selected_channel(ImageType.RAW)

    assert training_overrides[
        "target_col1_channel"
    ] == training_model.get_selected_channel(ImageType.SEG1)

    assert training_overrides[
        "target_col2_channel"
    ] == training_model.get_selected_channel(ImageType.SEG2)


def test_get_training_overrides_iterative_training(
    experiments_model: ExperimentsModel,
):
    # ARRANGE
    model: TrainingModel = TrainingModel(MainModel(), experiments_model)
    model.set_spatial_dims(3)
    model.set_images_directory("/path/to/images")
    model.set_selected_channel(ImageType.RAW, 1)
    model.set_selected_channel(ImageType.SEG1, 2)
    model.set_selected_channel(ImageType.SEG2, 3)
    model.set_use_max_time(False)
    model.set_max_time(9992)
    model.set_patch_size([4, 4, 4])
    model.set_num_epochs(100)
    model.set_is_using_existing_model(True)
    model.set_existing_model("2_exp")

    cyto_overrides_manager: CytoDLOverridesManager = CytoDLOverridesManager(
        experiments_model, model
    )

    # ACT
    training_overrides: Dict[str, Union[str, int, float, bool, Dict]] = (
        cyto_overrides_manager.get_training_overrides()
    )

    # ASSERT
    # Stuff related to iterative training:
    assert training_overrides["checkpoint.weights_only"]
    assert not training_overrides["checkpoint.strict"]
    assert training_overrides["checkpoint.ckpt_path"] == str(
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "main"
        / "experiments_home"
        / "2_exp"
        / "checkpoints"
        / "1.ckpt"
    )
    assert (
        training_overrides["paths.output_dir"]
        == f"{experiments_model.get_user_experiments_path()}/{experiments_model.get_experiment_name()}"
    )
    assert "model._aux.filters" not in training_overrides.keys()

    # Rest of the overrides should be the same as the previous test (regular training)
    assert training_overrides["spatial_dims"] == model.get_spatial_dims()

    # the experiments model fixture is set to the experiment @ _tests/experiments_home/one_ckpt_exp,
    # for which the last 'best' checkpoint is epoch_000.ckpt. Since we are currently at the first epoch checkpoint,
    # in order to run training_model.get_num_epochs() more epochs, we need to increment by 1
    assert training_overrides["trainer.max_epochs"] == model.get_num_epochs()

    assert training_overrides["data.path"] == str(model.get_images_directory())

    assert (
        training_overrides["data._aux.patch_shape"] == model.get_patch_size()
    )

    assert training_overrides["input_channel"] == model.get_selected_channel(
        ImageType.RAW
    )

    assert training_overrides[
        "target_col1_channel"
    ] == model.get_selected_channel(ImageType.SEG1)

    assert training_overrides[
        "target_col2_channel"
    ] == model.get_selected_channel(ImageType.SEG2)


def test_get_training_overrides_2d_spatial_dims(experiments_model) -> None:
    # Arrange
    model: TrainingModel = TrainingModel(MainModel(), experiments_model)
    model.set_experiment_type("segmentation")
    model.set_images_directory("/path/to/images")
    model.set_use_max_time(True)
    model.set_selected_channel(ImageType.RAW, 0)
    model.set_selected_channel(ImageType.SEG1, 0)
    model.set_selected_channel(
        ImageType.SEG2, 0
    )  # set to 0 if 2d image loaded
    model.set_max_time(9992)
    model.set_num_epochs(100)
    model.set_model_size("medium")

    model.set_spatial_dims(2)
    model.set_patch_size([4, 8])
    cyto_overrides_manager: CytoDLOverridesManager = CytoDLOverridesManager(
        experiments_model, model
    )

    # Act
    training_overrides: Dict[str, Union[str, int, float, bool, Dict]] = (
        cyto_overrides_manager.get_training_overrides()
    )

    # Assert
    assert len(training_overrides["data._aux.patch_shape"]) == 2
    assert training_overrides["data._aux.patch_shape"] == [4, 8]
    assert training_overrides["input_channel"] == 0


def test_max_epochs_no_existing_ckpt(
    experiments_model: ExperimentsModel, training_model: TrainingModel
):
    experiments_model.apply_experiment_name("0_exp")
    cyto_overrides_manager: CytoDLOverridesManager = CytoDLOverridesManager(
        experiments_model, training_model
    )

    training_overrides: Dict[str, Union[str, int, float, bool, Dict]] = (
        cyto_overrides_manager.get_training_overrides()
    )

    # the experiments model fixture is set to the experiment @ _tests/experiments_home/0_exp,
    # for which there are no existing checkpoints. So, we expect the max epoch to be equal to what the
    # user has entered in the field
    assert (
        training_overrides["trainer.max_epochs"]
        == training_model.get_num_epochs()
    )



from napari.layers import Labels
import numpy as np

from allencell_ml_segmenter.main.viewer import Viewer


def test_clear_binary_map_from_layer() -> None:
    layer: Labels = Labels(np.ones((10, 10, 10), dtype=bool))

    Viewer.clear_binary_map_from_layer(Viewer, layer)

    assert np.all(layer.data == 0)  # assert all values are now zeros



from enum import Enum
from pathlib import Path
from typing import List, Optional

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.publisher import Publisher
from allencell_ml_segmenter.core.file_input_model import FileInputModel


class PredictionModel(Publisher):
    """
    Stores state relevant to prediction processes.
    """

    def __init__(self) -> None:
        super().__init__()
        # state related to ModelInputWidget
        self._preprocessing_method: Optional[str] = None
        self._postprocessing_method: Optional[str] = None
        self._postprocessing_simple_threshold: Optional[float] = None
        self._postprocessing_auto_threshold: Optional[str] = None

        # This is initialized as None, and set when the input data is processed during pre-processing
        # If it is none after Event.ACTION_PREDICTION_SETUP is dispatched, the csv for
        # prediction was not generated and prediction cannot continue.
        self.total_num_images: Optional[int] = None

    def dispatch_prediction(self) -> None:
        """
        Dispatches an event to start prediction
        """
        # Shoots off a prediction run
        self.dispatch(Event.PROCESS_PREDICTION)

    def dispatch_prediction_get_image_paths_from_napari(self) -> None:
        # Does some pre-configuring if needed for prediction runs
        self.dispatch(Event.ACTION_PREDICTION_GET_IMAGE_PATHS_FROM_NAPARI)

    def dispatch_prediction_setup(self) -> None:
        # Does some pre-configuring if needed for prediction runs
        self.dispatch(Event.ACTION_PREDICTION_SETUP)

    def set_total_num_images(self, total: Optional[int]) -> None:
        self.total_num_images = total

    def get_total_num_images(self) -> Optional[int]:
        return self.total_num_images



from pathlib import Path
from typing import Optional
from xml.etree import ElementTree
from xml.etree.ElementTree import Element
import requests  # type: ignore

from allencell_ml_segmenter.utils.s3.s3_available_model import AvailableModel
from allencell_ml_segmenter.utils.s3.s3_request_exception import (
    S3RequestException,
)
from allencell_ml_segmenter.utils.s3.s3_bucket_constants import XML_NAMESPACES


class S3ModelBucket:
    def __init__(self, bucket_endpoint: str, path_to_save_models: Path):
        self._bucket_endpoint: str = bucket_endpoint
        self._path_to_save_models: Path = path_to_save_models

    def get_available_models(self) -> dict[str, AvailableModel]:
        """
        Get a dict of the filenames of available models on s3.

        Uses a http request to get all available models on s3. This function assumes all models will be in .zip files
        (and ignores all files that are not zip files). Any errors with the request throws a S3RequestException.

        Returns dictionary containing model name keys, and AvailableModels values.
        """
        # request parameter to list all objects in bucket
        req_params: str = "list-type=2"

        # request all bucket objects
        response: requests.Response = requests.get(
            f"{self._bucket_endpoint}?{req_params}"
        )

        if response.status_code == 200:
            # parse XML response with key
            model_names: set[str] = (
                self._parse_s3_xml_filelist_for_model_names(response.content)
            )
            # Create Dict of available Models
            # Where key- model_name and value- AvailableModels object (which stores the endpoint to download each model)
            available_models_dict: dict[str, AvailableModel] = {}
            for model_name in model_names:
                available_models_dict[model_name.split(".")[0]] = (
                    AvailableModel(
                        model_name,
                        self._bucket_endpoint,
                        self._path_to_save_models,
                    )
                )
            return available_models_dict
        else:
            # There was an issue with s3:ListBucket
            raise S3RequestException(
                f"S3 Download failed with status code {response.status_code}. {response.text}"
            )

    def _parse_s3_xml_filelist_for_model_names(
        self, response: bytes
    ) -> set[str]:
        """
        Parse XML response from s3 containing list of available objects for model names.
        Returns list of available models (which are .zip files) from s3's XML response.
        """
        # parse XML response
        xml_root: Element = ElementTree.fromstring(response)
        available: set[str] = set()
        # find all XML elements which match :param element_name: and get its "Key" element
        # which contains the object's filename
        all_s3_objects: list[str] = []
        for xml_content in xml_root.findall("aws_s3:Contents", XML_NAMESPACES):
            elem: Optional[Element] = xml_content.find(
                "aws_s3:Key", XML_NAMESPACES
            )
            if elem is not None and elem.text is not None:
                all_s3_objects.append(elem.text)

        for file in all_s3_objects:
            # append file name if it is a zip file- these are models stored on s3
            if file.endswith(".zip"):
                # S3 allows duplicate files- ensure we have unique names for all models on s3.
                if file in available:
                    raise ValueError(
                        f"S3 bucket {self._bucket_endpoint} contains duplicate models for name: {file}"
                    )
                available.add(file)
        return available



import numpy as np
from dataclasses import dataclass
from pathlib import Path
from typing import Optional


@dataclass
class CurationRecord:
    raw_file: Path
    seg1: Path
    seg2: Optional[Path]
    excluding_mask: Optional[np.ndarray]
    merging_mask: Optional[np.ndarray]
    base_image: Optional[str]
    to_use: bool



from pathlib import Path
from allencell_ml_segmenter.config.i_user_settings import IUserSettings
from qtpy.QtWidgets import QWidget


class FakeUserSettings(IUserSettings):
    def __init__(
        self,
        cyto_dl_home_path=None,
        user_experiments_path=None,
        init_prompt_response: Path = None,
        change_prompt_response: Path = None,
    ):
        self.cyto_dl_home_path = cyto_dl_home_path
        self.user_experiments_path = user_experiments_path
        self.prompt_response: Path = init_prompt_response
        self.change_prompt_response: Path = change_prompt_response

    def get_cyto_dl_home_path(self) -> Path:
        return self.cyto_dl_home_path

    def set_cyto_dl_home_path(self, path: Path) -> Path:
        self.cyto_dl_home_path = path

    def get_user_experiments_path(self) -> Path:
        return self.user_experiments_path

    def set_user_experiments_path(self, path: str):
        self.user_experiments_path = path

    def prompt_for_user_experiments_home(self, parent: QWidget):
        if self.prompt_response:
            self.set_user_experiments_path(Path(self.prompt_response))

    def display_change_user_experiments_home(self, parent: QWidget):
        if self.change_prompt_response:
            self.set_user_experiments_path(Path(self.prompt_response))



from pathlib import Path
import responses
import pytest
from pytest import MonkeyPatch

import allencell_ml_segmenter
from allencell_ml_segmenter.core.dialog_box import DialogBox
from allencell_ml_segmenter.utils.s3.s3_available_model import AvailableModel
from allencell_ml_segmenter.utils.s3.s3_request_exception import (
    S3RequestException,
)
from allencell_ml_segmenter.utils.zip_file.fake_zip_file_manager import (
    FakeZipFileManager,
)


@responses.activate
def test_download_model_and_unzip_sucessful_request() -> None:
    # fake data
    fake_content: str = "fake_content abcde"
    fake_url: str = "http://test.com/abc"
    fake_model_file_name: str = "some_new_model.zip"
    responses.add(
        **{
            "method": responses.GET,
            "url": f"{fake_url}/{fake_model_file_name}",
            "body": fake_content,
            "status": 200,
            "content_type": "application/zip",
            "adding_headers": {"X-Foo": "Bar"},
        }
    )
    # Test path to save zip to
    test_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "zip_files"
    )
    fake_zip_file_manager: FakeZipFileManager = FakeZipFileManager()
    available_model: AvailableModel = AvailableModel(
        fake_model_file_name, fake_url, test_path, fake_zip_file_manager
    )

    # Act
    available_model.download_model_and_unzip()

    # Assert
    assert fake_zip_file_manager.written_zip_files[
        test_path / fake_model_file_name
    ] == bytes(fake_content, "utf-8")
    assert (
        test_path / fake_model_file_name
        in fake_zip_file_manager.unzipped_files
    )


@responses.activate
def test_download_model_and_unzip_bad_request() -> None:
    # ARRANGE
    test_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "zip_files"
        / "test_zip.zip"
    )
    fake_url: str = "https://testurl.com/test_url"
    fake_zip_file_manager: FakeZipFileManager = FakeZipFileManager()
    available_model: AvailableModel = AvailableModel(
        "abc", fake_url, test_path, fake_zip_file_manager
    )
    responses.add(
        **{
            "method": responses.GET,
            "url": f"{fake_url}/abc",
            "body": '{"error": "error_reason"}',
            "status": 400,
            "content_type": "application/zip",
            "adding_headers": {"X-Foo": "Bar"},
        }
    )

    # Act/Assert
    with pytest.raises(S3RequestException):
        available_model.download_model_and_unzip()


@responses.activate
def test_download_model_and_unzip_model_exists_overwrite(
    monkeypatch: MonkeyPatch,
) -> None:
    # standard way to deal with modal dialogs: https://pytest-qt.readthedocs.io/en/latest/note_dialogs.html
    monkeypatch.setattr(DialogBox, "exec", lambda *args: 0)
    monkeypatch.setattr(
        DialogBox, "get_selection", lambda x: True
    )  # mimic user clicks ovewrite
    # fake data
    fake_content: str = "fake_content abcde"
    fake_url: str = "http://test.com/abc"
    # this is an existing "model" name in the test files directory
    fake_model_file_name: str = "test_zip.zip"
    responses.add(
        **{
            "method": responses.GET,
            "url": f"{fake_url}/{fake_model_file_name}",
            "body": fake_content,
            "status": 200,
            "content_type": "application/zip",
            "adding_headers": {"X-Foo": "Bar"},
        }
    )
    # Test path to save zip to
    test_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "zip_files"
    )
    fake_zip_file_manager: FakeZipFileManager = FakeZipFileManager()
    available_model: AvailableModel = AvailableModel(
        fake_model_file_name, fake_url, test_path, fake_zip_file_manager
    )

    # Act
    available_model.download_model_and_unzip()

    # Assert
    assert fake_zip_file_manager.written_zip_files[
        test_path / fake_model_file_name
    ] == bytes(fake_content, "utf-8")
    assert (
        test_path / fake_model_file_name
        in fake_zip_file_manager.unzipped_files
    )


@responses.activate
def test_download_model_and_unzip_model_exists_do_not_overwrite(
    monkeypatch: MonkeyPatch,
) -> None:
    # standard way to deal with modal dialogs: https://pytest-qt.readthedocs.io/en/latest/note_dialogs.html
    monkeypatch.setattr(DialogBox, "exec", lambda *args: 0)
    monkeypatch.setattr(
        DialogBox, "get_selection", lambda x: False
    )  # mimic user clicking dont overwrite
    # fake data
    fake_content: str = "fake_content abcde"
    fake_url: str = "http://test.com/abc"
    # this is an existing "model" name in the test files directory
    fake_model_file_name: str = "test_zip.zip"
    responses.add(
        **{
            "method": responses.GET,
            "url": f"{fake_url}/{fake_model_file_name}",
            "body": fake_content,
            "status": 200,
            "content_type": "application/zip",
            "adding_headers": {"X-Foo": "Bar"},
        }
    )
    # Test path to save zip to
    test_path: Path = (
        Path(allencell_ml_segmenter.__file__).parent
        / "_tests"
        / "test_files"
        / "zip_files"
    )
    fake_zip_file_manager: FakeZipFileManager = FakeZipFileManager()
    available_model: AvailableModel = AvailableModel(
        fake_model_file_name, fake_url, test_path, fake_zip_file_manager
    )

    # Act
    available_model.download_model_and_unzip()

    # Assert
    assert len(fake_zip_file_manager.written_zip_files) == 0
    assert len(fake_zip_file_manager.unzipped_files) == 0



from abc import ABC, abstractmethod
from pathlib import Path


class IZipFileManager(ABC):
    """
    ZipFileManager interface
    """

    @abstractmethod
    def write_zip_file(self, path: Path, contents: bytes) -> None:
        """
        Write zip file contents :param contents: to :param path:
        """
        pass

    @abstractmethod
    def unzip_zipped_file_and_delete_zip(self, path_to_zipped: Path) -> None:
        """
        Extract a zipped file to the same directory it is in (:param path_to_zipped: 's parent),
        and delete the original zip file at :param path_to_zipped:
        """
        pass



import pytest
from pathlib import Path

from qtpy.QtCore import Qt

from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter._tests.fakes.fake_subscriber import FakeSubscriber
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.thresholding.thresholding_model import (
    ThresholdingModel,
)
from allencell_ml_segmenter.core.file_input_model import (
    InputMode,
)
from allencell_ml_segmenter._tests.fakes.fake_viewer import FakeViewer
from allencell_ml_segmenter.thresholding.thresholding_view import (
    ThresholdingView,
)


@pytest.fixture
def main_model() -> MainModel:
    return MainModel()


@pytest.fixture
def thresholding_model(tmp_path: Path) -> ThresholdingModel:
    model = ThresholdingModel()
    model.set_thresholding_value(128)
    model.set_output_directory(tmp_path / "output")
    model.set_input_image_path(tmp_path / "input")
    model.set_input_mode(InputMode.FROM_PATH)
    return model


@pytest.fixture
def experiments_model() -> FakeExperimentsModel:
    return FakeExperimentsModel()


@pytest.fixture
def viewer() -> FakeViewer:
    return FakeViewer()


@pytest.fixture
def thresholding_view(
    main_model,
    thresholding_model,
    experiments_model,
    viewer,
    qtbot,
):
    view = ThresholdingView(
        main_model,
        thresholding_model,
        experiments_model,
        viewer,
    )
    qtbot.addWidget(view)
    return view


def test_model_updates_on_slider_release(
    thresholding_view, thresholding_model
):
    # this tests to see if the model updates when the slider releases
    # Arrange
    initial_value: int = thresholding_model.get_thresholding_value()

    # Act
    thresholding_view._threshold_value_slider.setValue(111)
    thresholding_view._threshold_value_slider.sliderReleased.emit()

    # Assert
    assert thresholding_model.get_thresholding_value() == 111
    assert thresholding_model.get_thresholding_value() != initial_value


def test_model_updates_on_spinbox_editing_finished(
    thresholding_view, thresholding_model
):
    # this tests to see if the model updates when the spinbox is edited
    # Arrange
    initial_value = thresholding_model.get_thresholding_value()

    # Act
    new_value = 122
    thresholding_view._threshold_value_spinbox.setValue(new_value)
    thresholding_view._threshold_value_spinbox.editingFinished.emit()

    # Assert: Model value should update to match the spinbox's new value
    assert thresholding_model.get_thresholding_value() == new_value
    assert thresholding_model.get_thresholding_value() != initial_value


def test_update_spinbox_from_slider(thresholding_view, qtbot):
    # This tests to see if the spinbox updates live along with the slider
    # Act: dragging
    for new_value in range(
        60, 101, 10
    ):  # Simulate dragging from 60 to 100 in steps
        thresholding_view._threshold_value_slider.setValue(new_value)
        qtbot.wait(10)  # small delay

        # Assert that spinbox always stays updated
        assert thresholding_view._threshold_value_spinbox.value() == new_value

    # Act: clicking a value
    qtbot.mouseClick(thresholding_view._threshold_value_slider, Qt.LeftButton)
    clicked_value = 120
    thresholding_view._threshold_value_slider.setValue(clicked_value)

    # Assert that spinbox updated correctly
    assert thresholding_view._threshold_value_spinbox.value() == clicked_value


def test_update_slider_from_spinbox(thresholding_view, qtbot):
    # This tests to see if the slider updates live when the spinbox is changed
    # Act
    new_value = 100
    thresholding_view._threshold_value_spinbox.setValue(new_value)
    qtbot.keyPress(
        thresholding_view._threshold_value_spinbox, Qt.Key_Enter
    )  # Simulate user pressing "Enter"

    # Assert
    assert thresholding_view._threshold_value_slider.value() == new_value


def test_update_state_from_radios(thresholding_view, thresholding_model):
    # This tests if ui/model updates correctly when the radio buttons are toggled
    # Arrange
    assert not thresholding_view._none_radio_button.isChecked()
    assert not thresholding_view._specific_value_radio_button.isChecked()
    assert not thresholding_view._autothreshold_radio_button.isChecked()
    assert not thresholding_view._apply_save_button.isEnabled()
    assert not thresholding_view._threshold_value_slider.isEnabled()
    assert not thresholding_view._threshold_value_spinbox.isEnabled()

    # Act
    thresholding_view._specific_value_radio_button.setChecked(True)
    thresholding_view._update_state_from_radios()

    # Assert
    assert thresholding_model.is_threshold_enabled()
    assert not thresholding_model.is_autothresholding_enabled()
    assert thresholding_view._apply_save_button.isEnabled()
    assert thresholding_view._threshold_value_slider.isEnabled()
    assert thresholding_view._threshold_value_spinbox.isEnabled()

    thresholding_view._specific_value_radio_button.setChecked(False)
    thresholding_view._autothreshold_radio_button.setChecked(True)
    thresholding_view._update_state_from_radios()

    # Assert
    assert not thresholding_model.is_threshold_enabled()
    assert thresholding_model.is_autothresholding_enabled()
    assert thresholding_view._apply_save_button.isEnabled()
    assert thresholding_view._autothreshold_method_combo.isEnabled()
    assert not thresholding_view._threshold_value_slider.isEnabled()
    assert not thresholding_view._threshold_value_spinbox.isEnabled()


def test_check_able_to_threshold_valid(
    main_model, experiments_model, viewer, tmp_path
):
    thresholding_model: ThresholdingModel = ThresholdingModel()
    thresholding_model.set_threshold_enabled(True)
    thresholding_model.set_thresholding_value(100)
    thresholding_model.set_output_directory(tmp_path / "output")
    thresholding_model.set_input_image_path(tmp_path / "input")
    thresholding_model.set_input_mode(InputMode.FROM_PATH)

    thresholding_view: ThresholdingView = ThresholdingView(
        main_model,
        thresholding_model,
        experiments_model,
        viewer,
    )

    assert thresholding_view._check_able_to_threshold()


def test_check_able_to_threshold_no_output_dir(
    main_model, experiments_model, viewer
):
    thresholding_model: ThresholdingModel = ThresholdingModel()
    thresholding_model.set_threshold_enabled(True)
    thresholding_model.set_thresholding_value(100)
    thresholding_model.set_input_mode(InputMode.FROM_PATH)
    thresholding_model.set_input_image_path(Path("fake_path"))
    thresholding_view: ThresholdingView = ThresholdingView(
        main_model,
        thresholding_model,
        experiments_model,
        viewer,
    )

    assert not thresholding_view._check_able_to_threshold()


def test_check_able_to_threshold_no_input_dir(
    main_model, experiments_model, viewer
):
    thresholding_model: ThresholdingModel = ThresholdingModel()
    thresholding_model.set_threshold_enabled(True)
    thresholding_model.set_thresholding_value(100)
    thresholding_model.set_input_mode(InputMode.FROM_PATH)
    thresholding_model.set_output_directory(Path("fake_path"))
    thresholding_view: ThresholdingView = ThresholdingView(
        main_model,
        thresholding_model,
        experiments_model,
        viewer,
    )

    assert not thresholding_view._check_able_to_threshold()


def test_check_able_to_threshold_no_input_method(
    main_model, experiments_model, viewer
):
    thresholding_model: ThresholdingModel = ThresholdingModel()
    thresholding_model.set_threshold_enabled(True)
    thresholding_model.set_thresholding_value(100)
    thresholding_model.set_input_image_path(Path("fake_path"))
    thresholding_model.set_output_directory(Path("fake_path"))
    thresholding_view: ThresholdingView = ThresholdingView(
        main_model,
        thresholding_model,
        experiments_model,
        viewer,
    )

    assert not thresholding_view._check_able_to_threshold()


def check_button_press_dispatches_event(
    thresholding_view, thresholding_model, qtbot
):
    # arrange
    fake_subscriber: FakeSubscriber = FakeSubscriber()
    thresholding_model.subscribe(
        Event.ACTION_SAVE_THRESHOLDING_IMAGES,
        fake_subscriber,
        fake_subscriber.handle,
    )

    # act
    qtbot.mouseClick(thresholding_view._apply_save_button, Qt.LeftButton)

    # assert that event was dispatched
    assert fake_subscriber.handled[Event.ACTION_SAVE_THRESHOLDING_IMAGES]



from pathlib import Path
from typing import Optional, Dict

from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.core.file_input_model import FileInputModel
from allencell_ml_segmenter.core.channel_extraction import (
    ChannelExtractionThread,
    get_img_path_from_csv,
)
from allencell_ml_segmenter.utils.file_utils import FileUtils
from napari.utils.notifications import show_error  # type: ignore


class ModelFileService(Subscriber):
    """
    Parses the chosen model file to extract the preprocessing method.
    """

    def __init__(self, model: FileInputModel):
        super().__init__()
        self._model: FileInputModel = model
        self._current_thread: Optional[ChannelExtractionThread] = None
        self._running_threads: Dict[int, ChannelExtractionThread] = {}
        self._threads_created = 0

        self._model.subscribe(
            Event.ACTION_FILEINPUT_EXTRACT_CHANNELS,
            self,
            lambda e: self._initiate_channel_extraction(),
        )

    def handle_event(self, event: Event) -> None:
        pass

    def stop_channel_extraction(self) -> None:
        if self._current_thread and self._current_thread.isRunning():
            self._current_thread.requestInterruption()

    def _get_img_path_from_model(self) -> Path:
        """
        Returns path of an image to be predicted from the prediction model. If model's
        input_image_path is set, will attempt to infer image from that field;
        otherwise will attempt to infer image from model's selected_paths field
        (the images currently checked in the prediction widget). Throws ValueError
        if both fields uninitialized.
        """
        path: Optional[Path] = self._model.get_input_image_path()
        img_path: Optional[Path] = None
        if not path:  # using viewer input method
            paths: Optional[list[Path]] = self._model.get_selected_paths()
            if not paths or len(paths) <= 0:
                raise ValueError(
                    "expected input_image_path or selected_paths to be initialized and non-empty"
                )
            img_path = Path(paths[0])
        elif path.is_dir():
            img_path = FileUtils.get_img_path_from_folder(path)
        elif path.suffix == ".csv":
            img_path = get_img_path_from_csv(path)
        else:
            raise ValueError(f"unrecognized input image path in model: {path}")

        return img_path

    def _initiate_channel_extraction(self) -> None:
        img_path: Path = self._get_img_path_from_model()

        self.stop_channel_extraction()

        self._current_thread = ChannelExtractionThread(img_path)
        thread_id: int = self._threads_created
        self._threads_created += 1
        self._running_threads[thread_id] = self._current_thread

        self._current_thread.channels_ready.connect(
            self._model.set_max_channels
        )
        self._current_thread.task_failed.connect(
            self._on_channel_extraction_failed
        )
        self._current_thread.finished.connect(
            lambda: self._on_channel_extraction_finished(thread_id)
        )
        self._current_thread.start()

    def _on_channel_extraction_finished(self, thread_id: int) -> None:
        self._running_threads.pop(thread_id)

    def _on_channel_extraction_failed(self, err: Exception) -> None:
        # setting max channels to none will result in the spinner stopping
        # but no channel selection being available
        self._model.set_max_channels(None)
        show_error(f"Channel extraction failed: {err}")



from pathlib import Path
from dataclasses import dataclass

import pytest
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter.curation.curation_model import (
    CurationModel,
    ImageType,
    CurationView,
)
from allencell_ml_segmenter.curation.curation_service import (
    CurationService,
)
from allencell_ml_segmenter.core.image_data_extractor import (
    FakeImageDataExtractor,
)
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter.utils.file_writer import FakeFileWriter
import allencell_ml_segmenter
from allencell_ml_segmenter.main.main_model import MainModel


FAKE_CHANNEL_SELECTION_PATH: Path = Path("channel_sel")

IMG_DIR_PATH = (
    Path(allencell_ml_segmenter.__file__).parent
    / "_tests"
    / "test_files"
    / "img_folder"
)

IMG_DIR_FILES = [path for path in IMG_DIR_PATH.iterdir()]


@dataclass
class TestEnvironment:
    model: CurationModel
    service: CurationService
    file_writer: FakeFileWriter


@pytest.fixture
def test_env_input_view() -> TestEnvironment:
    exp_mod: FakeExperimentsModel = FakeExperimentsModel()
    exp_mod.apply_experiment_name("0_exp")
    model: CurationModel = CurationModel(exp_mod, MainModel())
    writer: FakeFileWriter = FakeFileWriter()
    service: CurationService = CurationService(
        model,
        exp_mod,
        img_data_extractor=FakeImageDataExtractor.global_instance(),
        file_writer=writer,
    )
    return TestEnvironment(model, service, writer)


@pytest.fixture
def test_env_main_view(
    test_env_input_view: TestEnvironment,
) -> TestEnvironment:
    model: CurationModel = test_env_input_view.model
    model.set_image_directory_paths(ImageType.RAW, IMG_DIR_FILES)
    model.set_image_directory_paths(ImageType.SEG1, IMG_DIR_FILES)
    model.set_image_directory_paths(ImageType.SEG2, IMG_DIR_FILES)
    model.set_selected_channel(ImageType.RAW, 0)
    model.set_selected_channel(ImageType.SEG1, 0)
    model.set_selected_channel(ImageType.SEG2, 0)

    model.set_current_view(CurationView.MAIN_VIEW)

    return test_env_input_view


# NOTE: the service tests follow this general format:
# 1. act on the model to trigger a signal which service should be connected to
# 2. wait for signal indicating that service work triggered by 1 is complete
# 3. verify that expected side effects (most likely to model state) of completed work in 2
#    are present


def test_service_reacts_to_set_raw_dir(
    qtbot: QtBot, test_env_input_view: TestEnvironment
) -> None:
    test_env: TestEnvironment = test_env_input_view
    # Assert (sanity check)
    assert test_env.model.get_channel_count(ImageType.RAW) is None
    assert test_env.model.get_image_directory_paths(ImageType.RAW) is None

    # Act
    with qtbot.waitSignal(test_env.model.channel_count_set):
        test_env.model.set_image_directory(ImageType.RAW, IMG_DIR_PATH)

    # Assert
    # we expect that when the raw directory is set, CurationService will extract the paths
    # and the number of channels from that directory
    assert test_env.model.get_channel_count(ImageType.RAW) is not None
    for file in test_env.model.get_image_directory_paths(ImageType.RAW):
        assert file in IMG_DIR_FILES


def test_service_reacts_to_set_seg1_dir(
    qtbot: QtBot, test_env_input_view: TestEnvironment
) -> None:
    test_env: TestEnvironment = test_env_input_view
    # Assert (sanity check)
    assert test_env.model.get_channel_count(ImageType.SEG1) is None
    assert test_env.model.get_image_directory_paths(ImageType.SEG1) is None

    # Act
    with qtbot.waitSignal(test_env.model.channel_count_set):
        test_env.model.set_image_directory(ImageType.SEG1, IMG_DIR_PATH)

    # Assert
    assert test_env.model.get_channel_count(ImageType.SEG1) is not None
    for file in test_env.model.get_image_directory_paths(ImageType.SEG1):
        assert file in IMG_DIR_FILES


def test_service_reacts_to_set_seg2_dir(
    qtbot: QtBot, test_env_input_view: TestEnvironment
) -> None:
    test_env: TestEnvironment = test_env_input_view
    # Assert (sanity check)
    assert test_env.model.get_channel_count(ImageType.SEG2) is None
    assert test_env.model.get_image_directory_paths(ImageType.SEG2) is None

    # Act
    with qtbot.waitSignal(test_env.model.channel_count_set):
        test_env.model.set_image_directory(ImageType.SEG2, IMG_DIR_PATH)

    # Assert
    assert test_env.model.get_channel_count(ImageType.SEG2) is not None
    for file in test_env.model.get_image_directory_paths(ImageType.SEG2):
        assert file in IMG_DIR_FILES


def test_service_reacts_to_cursor_moved(
    qtbot: QtBot, test_env_main_view: TestEnvironment
) -> None:
    test_env: TestEnvironment = test_env_main_view
    # Assert (sanity check)
    assert test_env.model.get_curr_image_data(ImageType.RAW) is None
    assert test_env.model.get_curr_image_data(ImageType.SEG1) is None
    assert test_env.model.get_curr_image_data(ImageType.SEG2) is None

    # Act
    with qtbot.waitSignal(test_env.model.image_loading_finished):
        test_env.model.start_loading_images()

    # Assert
    assert not test_env.model.is_waiting_for_images()
    assert test_env.model.get_curr_image_data(ImageType.RAW) is not None
    assert test_env.model.get_curr_image_data(ImageType.SEG1) is not None
    assert test_env.model.get_curr_image_data(ImageType.SEG2) is not None


def test_service_reacts_to_save_csv(
    qtbot: QtBot, test_env_main_view: TestEnvironment
) -> None:
    test_env: TestEnvironment = test_env_main_view

    # Arrange
    with qtbot.waitSignal(test_env.model.image_loading_finished):
        test_env.model.start_loading_images()

    # get to the fourth image so that we can save csv
    with qtbot.waitSignal(test_env.model.image_loading_finished):
        test_env.model.next_image()
    with qtbot.waitSignal(test_env.model.image_loading_finished):
        test_env.model.next_image()
    with qtbot.waitSignal(test_env.model.image_loading_finished):
        test_env.model.next_image()

    # Assert (sanity check)
    assert len(test_env.file_writer.csv_state) == 0

    # Act
    with qtbot.waitSignal(test_env.model.saved_to_disk):
        test_env.model.save_curr_curation_record_to_disk()

    # Assert
    assert len(test_env.file_writer.csv_state) > 0



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Segmenter ML Plugin for napari"
copyright = "2024, Allen Institute for Cell Science"
author = "Segmenter ML plugin team"
release = "1.0.1rc0"


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinxext.opengraph",
    "sphinx_inline_tabs",
    "sphinx_copybutton",
    "myst_parser",
    "sphinx_togglebutton",
    "sphinx_design",
]

myst_enable_extensions = [
    "attrs_inline",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_image",
    "linkify",
    "substitution",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


def setup(app):
    app.add_css_file("theme_1_TD.css")


html_theme = "furo"
html_title = "Segmenter ML - User Guide"
html_static_path = ["_static"]
html_theme_options = {
    "light_logo": "images/thumb_segmenterML_2_onLight.png",
    "dark_logo": "images/thumb_segmenterML_2_onDark.png",
}


# html_css_files = ["custom.css"]



from .i_file_writer import IFileWriter
import numpy as np
from pathlib import Path
from typing import List, Dict, Any, Union


class FakeFileWriter(IFileWriter):
    _instance = None

    def __init__(self) -> None:
        # {path: saved_array}
        self.np_save_state: Dict[Path, np.ndarray] = {}

        # {path: {"open": T/F, "rows": [[header1, header2...], [col1, col2...]]}}
        self.csv_state: Dict[Path, Dict[str, Any]] = {}

        # {path: json-like-obj}
        self.json_state: dict[Path, Union[list, dict]] = {}

    def np_save(self, path: Path, arr: np.ndarray) -> None:
        """
        Saves :param arr: to :param path:
        """
        self.np_save_state[path.resolve()] = arr

    def csv_open_write_mode(self, path: Path) -> None:
        """
        Opens a CSV at :param path: in write mode. Can only call csv_write_row
        on the path after running this method. Must call csv_close with the same
        path at some point after calling this method.
        """
        path = path.resolve()
        if path in self.csv_state and self.csv_state[path]["open"]:
            raise RuntimeError(f"{path} already open")
        self.csv_state[path] = {"open": True, "rows": []}

    def csv_write_row(self, path: Path, row: List[str]) -> None:
        """
        Appends :param row: to the open CSV at :param path:
        """
        path = path.resolve()
        if path not in self.csv_state or not self.csv_state[path]["open"]:
            raise RuntimeError(
                f"{path} must be opened with csv_open_write_mode before writing"
            )
        self.csv_state[path]["rows"].append(row)

    def csv_close(self, path: Path) -> None:
        """
        Closes the open CSV at :param path:
        """
        path = path.resolve()
        if path in self.csv_state:
            self.csv_state[path]["open"] = False

    def write_json(self, json_like_obj: Union[list, dict], path: Path) -> None:
        self.json_state[path] = json_like_obj



from napari.utils.notifications import show_warning  # type: ignore
from pathlib import Path
from typing import Optional

from allencell_ml_segmenter.main.i_experiments_model import IExperimentsModel
from allencell_ml_segmenter.main.i_viewer import IViewer
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QLabel,
    QPushButton,
    QFrame,
    QVBoxLayout,
    QSizePolicy,
    QWidget,
    QGridLayout,
    QComboBox,
    QHBoxLayout,
    QRadioButton,
    QLineEdit,
    QCheckBox,
)
from allencell_ml_segmenter._style import Style
from allencell_ml_segmenter.core.event import Event
from allencell_ml_segmenter.core.view import View, MainWindow
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.training.image_selection_widget import (
    ImageSelectionWidget,
)
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ModelSize,
)
from allencell_ml_segmenter.training.patch_size_validator import (
    PatchSizeValidator,
    PATCH_SIZE_MULTIPLE_OF,
)
from allencell_ml_segmenter.widgets.label_with_hint_widget import LabelWithHint
from qtpy.QtGui import QIntValidator
from allencell_ml_segmenter.training.training_progress_tracker import (
    TrainingProgressTracker,
)
from allencell_ml_segmenter.core.info_dialog_box import InfoDialogBox
from allencell_ml_segmenter.utils.file_utils import FileUtils
from allencell_ml_segmenter.utils.experiment_utils import ExperimentUtils


class TrainingView(View, MainWindow):
    """
    Holds widgets pertinent to training processes - ImageSelectionWidget & ModelSelectionWidget.
    """

    def __init__(
        self,
        main_model: MainModel,
        experiments_model: IExperimentsModel,
        training_model: TrainingModel,
        viewer: IViewer,
    ):
        super().__init__()

        self._viewer: IViewer = viewer

        self._main_model: MainModel = main_model
        self._experiments_model: IExperimentsModel = experiments_model
        self._training_model: TrainingModel = training_model

        layout: QVBoxLayout = QVBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum
        )

        self._title: QLabel = QLabel("SEGMENTATION MODEL TRAINING", self)
        self._title.setObjectName("title")
        layout.addWidget(
            self._title, alignment=Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop  # type: ignore
        )

        # initialize constituent widgets
        self.image_selection_widget: ImageSelectionWidget = (
            ImageSelectionWidget(self._training_model, self._experiments_model)
        )
        self.image_selection_widget.setObjectName("imageSelection")

        # Dummy divs allow for easy alignment
        top_container: QVBoxLayout = QVBoxLayout()
        top_dummy: QFrame = QFrame()
        bottom_dummy: QFrame = QFrame()
        top_container.addWidget(self.image_selection_widget)
        top_dummy.setLayout(top_container)
        layout.addWidget(top_dummy)

        # bottom half
        bottom_grid_layout: QGridLayout = QGridLayout()

        # Existing model label
        existing_model_label: LabelWithHint = LabelWithHint(
            "Start from previous model"
        )
        existing_model_label.set_hint(
            "Select a previously trained model to pull model weights from"
        )
        bottom_grid_layout.addWidget(
            existing_model_label, 0, 0, alignment=Qt.AlignmentFlag.AlignTop
        )

        # Model Size Selection
        self._model_size_combo_box: QComboBox = QComboBox()
        self._model_size_combo_box.setObjectName("modelSizeComboBox")
        self._model_size_combo_box.setCurrentIndex(-1)
        self._model_size_combo_box.setPlaceholderText("Select an option")
        self._model_size_combo_box.addItems(
            [size.name.lower() for size in ModelSize]
        )
        self._model_size_combo_box.currentTextChanged.connect(
            lambda model_size: self._training_model.set_model_size(model_size)
        )

        # Existing model selection
        self.existing_model_dropdown: QComboBox = QComboBox()
        self.existing_model_dropdown.addItems(
            self._experiments_model.get_experiments()
        )
        self.existing_model_dropdown.currentIndexChanged.connect(
            lambda: self._training_model.set_existing_model(
                self.existing_model_dropdown.currentText()
            )
        )
        self.existing_model_dropdown.setEnabled(
            False
        )  # Disabled by default since no is the default
        # Yes and no radio buttons stacked on top of each other in a VBoxLayout
        existing_model_selection_layout: QVBoxLayout = QVBoxLayout()
        self.existing_model_no_radio: QRadioButton = QRadioButton("No")
        self.existing_model_no_radio.toggled.connect(
            self._existing_model_no_radio_slot
        )
        self.existing_model_no_radio.setChecked(True)  # No enabled by default
        existing_model_selection_layout.addWidget(self.existing_model_no_radio)
        self.existing_model_yes_radio: QRadioButton = QRadioButton("Yes")
        self.existing_model_yes_radio.toggled.connect(
            self._existing_model_yes_radio_slot
        )

        # have model selection dropdown in-line with yes button
        existing_model_yes_layout: QHBoxLayout = QHBoxLayout()
        existing_model_yes_layout.addWidget(self.existing_model_yes_radio)
        existing_model_yes_layout.addWidget(
            self.existing_model_dropdown, stretch=1
        )
        existing_model_selection_layout.addLayout(existing_model_yes_layout)
        bottom_grid_layout.addLayout(existing_model_selection_layout, 0, 1)

        patch_size_text_layout = QVBoxLayout()
        patch_size_text_layout.setSpacing(0)
        patch_size_label: LabelWithHint = LabelWithHint("Patch size")
        patch_size_label.set_hint(
            "Patch size to split images into during training. Should encompass the structure of interest and all dimensions should be evenly divisble by 4. If 2D, Z can be left blank."
        )
        patch_size_text_layout.addWidget(patch_size_label)
        guide_text: QLabel = QLabel(
            f"All values must be multiples of {PATCH_SIZE_MULTIPLE_OF}"
        )
        guide_text.setObjectName("subtext")
        patch_size_text_layout.addWidget(guide_text)
        bottom_grid_layout.addLayout(patch_size_text_layout, 1, 0)
        patch_size_entry_layout: QHBoxLayout = QHBoxLayout()

        # allow only integers for the linedits below
        patch_validator: PatchSizeValidator = PatchSizeValidator()

        self.z_patch_size: QLineEdit = QLineEdit()
        self.z_patch_size.setValidator(patch_validator)
        patch_size_entry_layout.addWidget(QLabel("Z:"))
        patch_size_entry_layout.addWidget(self.z_patch_size)

        self.y_patch_size: QLineEdit = QLineEdit()
        self.y_patch_size.setValidator(patch_validator)
        patch_size_entry_layout.addWidget(QLabel("Y:"))
        patch_size_entry_layout.addWidget(self.y_patch_size)

        self.x_patch_size: QLineEdit = QLineEdit()
        self.x_patch_size.setValidator(patch_validator)
        patch_size_entry_layout.addWidget(QLabel("X:"))
        patch_size_entry_layout.addWidget(self.x_patch_size)

        self._update_spatial_dims_boxes()
        bottom_grid_layout.addLayout(patch_size_entry_layout, 1, 1)

        model_size_label: LabelWithHint = LabelWithHint("Model size")
        model_size_label.set_hint(
            "Defines the complexity of the model - smaller models train faster, while large models train slower but may learn complex relationships better."
        )
        bottom_grid_layout.addWidget(model_size_label, 2, 0)
        bottom_grid_layout.addWidget(self._model_size_combo_box, 2, 1)
        num_epochs_label: LabelWithHint = LabelWithHint("Number of epochs")
        num_epochs_label.set_hint(
            "Each epoch is one complete pass through the entire training dataset. More epochs yields better performance at the cost of training time."
        )
        bottom_grid_layout.addWidget(num_epochs_label, 3, 0)

        self._num_epochs_input: QLineEdit = QLineEdit()
        int_validator: QIntValidator = QIntValidator()
        int_validator.setBottom(1)
        self._num_epochs_input.setValidator(int_validator)
        self._num_epochs_input.setPlaceholderText("1000")
        self._num_epochs_input.setObjectName("trainingStepInput")
        self._num_epochs_input.textChanged.connect(
            self._num_epochs_field_handler
        )
        bottom_grid_layout.addWidget(self._num_epochs_input, 3, 1)

        bottom_grid_layout.setColumnStretch(1, 8)
        bottom_grid_layout.setColumnStretch(0, 3)

        bottom_dummy.setLayout(bottom_grid_layout)
        layout.addWidget(bottom_dummy)

        self._train_btn: QPushButton = QPushButton("Start training")
        self._train_btn.setObjectName("trainBtn")
        layout.addWidget(self._train_btn)
        self._train_btn.clicked.connect(self.train_btn_handler)

        self._main_model.subscribe(
            Event.VIEW_SELECTION_TRAINING,
            self,
            lambda e: self._main_model.set_current_view(self),
        )
        self._training_model.signals.spatial_dims_set.connect(
            self._update_spatial_dims_boxes
        )

        # apply styling
        self.setStyleSheet(Style.get_stylesheet("training_view.qss"))

    def train_btn_handler(self) -> None:
        """
        Starts training process
        """
        if self._patch_size_ok():
            self.set_patch_size()
            num_epochs: Optional[int] = self._training_model.get_num_epochs()
            metrics_csv_path: Optional[Path] = (
                self._experiments_model.get_metrics_csv_path()
            )
            cache_dir: Optional[Path] = self._experiments_model.get_cache_dir()
            if metrics_csv_path is None or cache_dir is None:
                raise ValueError(
                    "Metrics CSV path and cache_dir cannot be None after training has started!"
                )
            progress_tracker: TrainingProgressTracker = (
                TrainingProgressTracker(
                    metrics_csv_path,
                    cache_dir,
                    num_epochs if num_epochs is not None else 0,
                    self._training_model.get_total_num_images(),
                    self._experiments_model.get_latest_metrics_csv_version()
                    + 1,
                )
            )
            self.startLongTaskWithProgressBar(progress_tracker)

    # Abstract methods from View implementations #######################

    def doWork(self) -> None:
        """
        Starts training process
        """
        self._training_model.dispatch_training()

    def getTypeOfWork(self) -> str:
        """
        Returns string representation of training process
        """
        return "Training"

    def showResults(self) -> None:
        # double check to see if a ckpt was generated
        exp_path: Optional[Path] = (
            self._experiments_model.get_user_experiments_path()
        )
        if exp_path is None:
            raise ValueError(
                "Experiments path should not be None after training complete."
            )
        exp_name: Optional[str] = self._experiments_model.get_experiment_name()
        if exp_name is None:
            raise ValueError(
                "Experiment name should not be None after training complete."
            )
        ckpt_generated: Optional[Path] = ExperimentUtils.get_best_ckpt(
            exp_path,
            exp_name,
        )
        if ckpt_generated is not None:
            # if model was successfully trained, get metrics to display
            csv_path: Optional[Path] = (
                self._experiments_model.get_latest_metrics_csv_path()
            )
            if csv_path is None:
                raise RuntimeError("Cannot get min loss from undefined csv")
            min_loss: Optional[float] = FileUtils.get_min_loss_from_csv(
                csv_path
            )
            if min_loss is None:
                raise RuntimeError("Cannot compute min loss")

            dialog_box = InfoDialogBox(
                "Training finished -- Final loss: {:.3f}".format(min_loss)
            )
            dialog_box.exec()  # this shows the dialog box
            self._main_model.training_complete()  # this dispatches the event that changes to prediction tab
        else:
            dialog_box = InfoDialogBox(
                "Training failed- no model was saved from this run."
            )
            dialog_box.exec()

    def _num_epochs_field_handler(self, num_epochs: str) -> None:
        self._training_model.set_num_epochs(int(num_epochs))

    def _max_time_checkbox_slot(self, checked: Qt.CheckState) -> None:
        """
        Triggered when the user selects the "time out after" _timeout_checkbox.
        Enables/disables interaction with the neighboring hour input based on checkstate.
        """
        if checked == Qt.CheckState.Checked:
            self._max_time_in_minutes_input.setEnabled(True)
            self._training_model.set_use_max_time(True)
        else:
            self._max_time_in_minutes_input.setEnabled(False)
            self._training_model.set_use_max_time(False)

    def _patch_size_ok(self) -> bool:
        """
        Validates that user has defined all needed patch sizes in the UI
        Returns True if all needed patch sizes were provided, False if not (and will show napari error message)
        """
        missing_patches: list[str] = []

        if (
            not self.z_patch_size.text()
            and self._training_model.get_spatial_dims() == 3
        ):
            missing_patches.append("Z")

        if not self.y_patch_size.text():
            missing_patches.append("Y")

        if not self.x_patch_size.text():
            missing_patches.append("X")

        if len(missing_patches) > 0:
            show_warning(
                f"Please define {missing_patches} patches before continuing."
            )
            return False

        return True

    def set_patch_size(self) -> None:
        self._training_model.set_patch_size(
            [
                int(self.z_patch_size.text()),
                int(self.y_patch_size.text()),
                int(self.x_patch_size.text()),
            ]
        )

    def focus_changed(self) -> None:
        self.image_selection_widget.set_inputs_csv()
        self._viewer.clear_layers()

    def _disable_model_size_combo_box(self) -> None:
        self._model_size_combo_box.setEnabled(False)
        self._model_size_combo_box.setCurrentIndex(
            -1
        )  # clear any selection, if any  # clear model

    def _enable_model_size_combo_box(self) -> None:
        self._model_size_combo_box.setEnabled(True)

    def _existing_model_no_radio_slot(self) -> None:
        # disable dropdown
        self.existing_model_dropdown.setEnabled(False)
        # reset dropdown and selected model
        self.existing_model_dropdown.setCurrentIndex(-1)
        self._training_model.set_is_using_existing_model(False)
        self._training_model.set_existing_model(None)
        self._enable_model_size_combo_box()

    def _existing_model_yes_radio_slot(self) -> None:
        # enable dropdown
        self._training_model.set_is_using_existing_model(True)
        self.existing_model_dropdown.setEnabled(True)
        self._disable_model_size_combo_box()

    def _update_spatial_dim_box(
        self, line_edit: QLineEdit, should_be_enabled: bool
    ) -> None:
        if should_be_enabled:
            line_edit.setEnabled(True)
            line_edit.setPlaceholderText(None)
        else:
            line_edit.setEnabled(False)
            line_edit.setPlaceholderText("N/A")

    def _update_spatial_dims_boxes(self) -> None:
        spatial_dims: Optional[int] = self._training_model.get_spatial_dims()
        self._update_spatial_dim_box(
            self.x_patch_size, spatial_dims is not None
        )
        self._update_spatial_dim_box(
            self.y_patch_size, spatial_dims is not None
        )
        self._update_spatial_dim_box(self.z_patch_size, spatial_dims == 3)



from qtpy.QtWidgets import QWidget, QSizePolicy, QHBoxLayout, QLabel
from qtpy.QtGui import QPixmap

from allencell_ml_segmenter.core.directories import Directories


class LabelWithHint(QWidget):
    """
    Compound widget with text label and question mark icon for clear access to tool tips.
    """

    def __init__(
        self, label_text: str = "", value_text: str = "", hint: str = ""
    ):
        super().__init__()

        self.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed
        )

        layout: QHBoxLayout = QHBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self._label: QLabel = QLabel("")
        self._label.setText(label_text)
        layout.addWidget(self._label)

        self._question_mark: QLabel = QLabel()
        self._question_mark.setPixmap(
            QPixmap(
                f"{str(Directories.get_assets_dir())}/icons/question-circle.svg"
            )
        )
        self._question_mark.setObjectName("questionMark")
        self._question_mark.setToolTip(hint)

        layout.addWidget(self._question_mark)
        self._question_mark.setVisible(bool(hint))

        self._label = QLabel(value_text)
        layout.addWidget(self._label)
        layout.addStretch(6)

    def set_label_text(self, text: str) -> None:
        """
        Sets the text of the label.
        """
        self._label.setText(text)

    def set_value_text(self, value_text: str) -> None:
        """
        Sets the text of the value.
        """
        self._label.setText(value_text)

    def set_hint(self, hint: str) -> None:
        """
        Sets the tooltip to be displayed when the question icon is hovered over.
        """
        self._question_mark.setToolTip(hint)
        self._question_mark.setVisible(bool(hint))

    def add_right_space(self, marg: int) -> None:
        """
        Sets margin-right such that the question mark icon is not cramped.
        """
        self._question_mark.setStyleSheet(
            "#questionMark {margin-right: " + str(marg) + "}"
        )



from pathlib import Path
from watchdog.observers.api import BaseObserver
from watchdog.observers import Observer
from allencell_ml_segmenter.core.progress_tracker import ProgressTracker
from allencell_ml_segmenter.prediction.file_write_event_handler import (
    FileWriteEventHandler,
)
from typing import Optional


class FileWriteProgressTracker(ProgressTracker):
    """
    A FileWriteProgressTracker measures progress by observing a folder in
    which cyto-dl will create images and incrementing progress when
    a new prediction image is created and placed in this folder.
    """

    def __init__(self, write_folder_path: Path, num_preds: int):
        """
        :param progress_folder_path: path to the output directory for predictions
        :param num_preds: total number of new pred files expected to be written to the folder
        """
        super().__init__(
            progress_minimum=0,
            progress_maximum=num_preds,
            label_text="Prediction progress",
        )

        if not write_folder_path.exists():
            write_folder_path.mkdir(parents=True)
        self._write_folder_path: Path = write_folder_path

        self._observer: Optional[BaseObserver] = None

    # override
    def start_tracker(self) -> None:
        self.stop_tracker()
        self._observer = Observer()
        event_handler: FileWriteEventHandler = FileWriteEventHandler(
            self.set_progress
        )
        self._observer.schedule(
            event_handler,
            path=str(self._write_folder_path.resolve()),
            recursive=False,
        )
        self._observer.start()

    # override
    def stop_tracker(self) -> None:
        if self._observer:
            self._observer.stop()



import pytest
from pathlib import Path
from unittest.mock import Mock
import numpy as np

from allencell_ml_segmenter.curation.curation_model import (
    CurationModel,
    CurationView,
    ImageType,
)
from allencell_ml_segmenter._tests.fakes.fake_experiments_model import (
    FakeExperimentsModel,
)
from allencell_ml_segmenter.main.main_model import MainModel
from allencell_ml_segmenter.core.image_data_extractor import ImageData
import numpy as np


FAKE_IMAGE_DATA: ImageData = ImageData(
    28, 28, 28, 1, np.zeros((28, 28, 28)), Path("fake")
)


@pytest.fixture
def curation_model() -> CurationModel:
    # returns curation model with view set to input view
    return CurationModel(FakeExperimentsModel(), MainModel())


@pytest.fixture
def curation_model_main_view(curation_model: CurationModel) -> CurationModel:
    # returns curation model configured to main view
    curation_model.set_image_directory_paths(
        ImageType.RAW, [Path("r1"), Path("r2"), Path("r3")]
    )
    curation_model.set_image_directory_paths(
        ImageType.SEG1, [Path("s11"), Path("s12"), Path("s13")]
    )
    curation_model.set_image_directory_paths(
        ImageType.SEG2, [Path("s21"), Path("s22"), Path("s23")]
    )
    curation_model.set_current_view(CurationView.MAIN_VIEW)
    return curation_model


@pytest.fixture
def curation_model_loading_started(
    curation_model_main_view: CurationModel,
) -> CurationModel:
    # returns curation model with first images done loading
    curation_model_main_view.start_loading_images()

    curation_model_main_view.set_curr_image_data(
        ImageType.RAW, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_curr_image_data(
        ImageType.SEG1, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_curr_image_data(
        ImageType.SEG2, FAKE_IMAGE_DATA
    )

    curation_model_main_view.set_next_image_data(
        ImageType.RAW, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_next_image_data(
        ImageType.SEG1, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_next_image_data(
        ImageType.SEG2, FAKE_IMAGE_DATA
    )
    return curation_model_main_view


def test_set_raw_directory(curation_model: CurationModel) -> None:
    # Arrange
    directory: Path = Path("fake_path")
    dir_set_slot: Mock = Mock()
    curation_model.image_directory_set.connect(dir_set_slot)

    # Act
    curation_model.set_image_directory(ImageType.RAW, directory)

    # Assert
    assert curation_model.get_image_directory(ImageType.RAW) == directory
    dir_set_slot.assert_called_once_with(ImageType.RAW)


def test_set_seg1_directory(curation_model: CurationModel) -> None:
    # Arrange
    directory: Path = Path("fake_path")
    dir_set_slot: Mock = Mock()
    curation_model.image_directory_set.connect(dir_set_slot)

    # Act
    curation_model.set_image_directory(ImageType.SEG1, directory)

    # Assert
    assert curation_model.get_image_directory(ImageType.SEG1) == directory
    dir_set_slot.assert_called_once_with(ImageType.SEG1)


def test_set_seg2_directory(curation_model: CurationModel) -> None:
    # Arrange
    directory: Path = Path("fake_path")
    dir_set_slot: Mock = Mock()
    curation_model.image_directory_set.connect(dir_set_slot)

    # Act
    curation_model.set_image_directory(ImageType.SEG2, directory)

    # Assert
    assert curation_model.get_image_directory(ImageType.SEG2) == directory
    dir_set_slot.assert_called_once_with(ImageType.SEG2)


def test_set_raw_image_channel_count(
    curation_model_loading_started: CurationModel,
) -> None:
    # Arrange
    channel_count_set_slot: Mock = Mock()
    curation_model_loading_started.channel_count_set.connect(
        channel_count_set_slot
    )

    # Act
    curation_model_loading_started.set_channel_count(ImageType.RAW, 4)

    # Assert
    assert curation_model_loading_started.get_channel_count(ImageType.RAW) == 4
    channel_count_set_slot.assert_called_once_with(ImageType.RAW)


def test_set_seg1_image_channel_count(
    curation_model_loading_started: CurationModel,
) -> None:
    # Arrange
    channel_count_set_slot: Mock = Mock()
    curation_model_loading_started.channel_count_set.connect(
        channel_count_set_slot
    )

    # Act
    curation_model_loading_started.set_channel_count(ImageType.SEG1, 5)

    # Assert
    assert (
        curation_model_loading_started.get_channel_count(ImageType.SEG1) == 5
    )
    channel_count_set_slot.assert_called_once_with(ImageType.SEG1)


def test_set_seg2_image_channel_count(
    curation_model_loading_started: CurationModel,
) -> None:
    # Arrange
    channel_count_set_slot: Mock = Mock()
    curation_model_loading_started.channel_count_set.connect(
        channel_count_set_slot
    )

    # Act
    curation_model_loading_started.set_channel_count(ImageType.SEG2, 6)

    # Assert
    assert (
        curation_model_loading_started.get_channel_count(ImageType.SEG2) == 6
    )
    channel_count_set_slot.assert_called_once_with(ImageType.SEG2)


def test_set_raw_channel(curation_model: CurationModel) -> None:
    # Arrange
    channel: int = 0

    # Act
    curation_model.set_selected_channel(ImageType.RAW, channel)

    # Assert
    assert curation_model.get_selected_channel(ImageType.RAW) == channel


def test_set_seg1_channel(curation_model: CurationModel) -> None:
    # Arrange
    channel: int = 1

    # Act
    curation_model.set_selected_channel(ImageType.SEG1, channel)

    # Assert
    assert curation_model.get_selected_channel(ImageType.SEG1) == channel


def test_set_seg2_channel(curation_model: CurationModel) -> None:
    # Arrange
    channel: int = 2

    # Act
    curation_model.set_selected_channel(ImageType.SEG2, channel)

    # Assert
    assert curation_model.get_selected_channel(ImageType.SEG2) == channel


def test_set_current_view_to_main_view(curation_model: CurationModel) -> None:
    # Arrange

    # expect paths to be set before changing view
    curation_model.set_image_directory_paths(
        ImageType.RAW, [Path("r1"), Path("r2"), Path("r3")]
    )
    curation_model.set_image_directory_paths(
        ImageType.SEG1, [Path("s11"), Path("s12"), Path("s13")]
    )
    curation_model.set_image_directory_paths(
        ImageType.SEG2, [Path("s21"), Path("s22"), Path("s23")]
    )

    view_changed_slot: Mock = Mock()
    curation_model.current_view_changed.connect(view_changed_slot)

    # Act
    curation_model.set_current_view(CurationView.MAIN_VIEW)

    # Assert
    assert curation_model.get_current_view() == CurationView.MAIN_VIEW

    view_changed_slot.assert_called_once()


def test_start_loading_images(curation_model_main_view: CurationModel) -> None:
    # Arrange
    cursor_moved_slot: Mock = Mock()
    curation_model_main_view.cursor_moved.connect(cursor_moved_slot)
    img_loading_finished_slot: Mock = Mock()
    curation_model_main_view.image_loading_finished.connect(
        img_loading_finished_slot
    )

    # Act
    curation_model_main_view.start_loading_images()

    # Assert
    cursor_moved_slot.assert_called_once()
    img_loading_finished_slot.assert_not_called()

    # Act (pretending to be curation service)
    curation_model_main_view.set_curr_image_data(
        ImageType.RAW, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_curr_image_data(
        ImageType.SEG1, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_curr_image_data(
        ImageType.SEG2, FAKE_IMAGE_DATA
    )

    curation_model_main_view.set_next_image_data(
        ImageType.RAW, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_next_image_data(
        ImageType.SEG1, FAKE_IMAGE_DATA
    )

    # Assert
    # still one unwritten image data, so expect img_loading_finished not to be emitted yet
    img_loading_finished_slot.assert_not_called()

    # Act
    curation_model_main_view.set_next_image_data(
        ImageType.SEG2, FAKE_IMAGE_DATA
    )

    # Assert
    img_loading_finished_slot.assert_called_once()
    assert (
        curation_model_main_view.get_curr_image_data(ImageType.RAW) is not None
    )
    assert (
        curation_model_main_view.get_curr_image_data(ImageType.SEG1)
        is not None
    )
    assert (
        curation_model_main_view.get_curr_image_data(ImageType.SEG2)
        is not None
    )


def test_next_image(curation_model_main_view: CurationModel) -> None:
    # Arrange
    cursor_moved_slot: Mock = Mock()
    curation_model_main_view.cursor_moved.connect(cursor_moved_slot)
    img_loading_finished_slot: Mock = Mock()
    curation_model_main_view.image_loading_finished.connect(
        img_loading_finished_slot
    )
    curation_model_main_view.start_loading_images()
    curation_model_main_view.set_curr_image_data(
        ImageType.RAW, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_curr_image_data(
        ImageType.SEG1, FAKE_IMAGE_DATA
    )

    curation_model_main_view.set_next_image_data(
        ImageType.RAW, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_next_image_data(
        ImageType.SEG1, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_next_image_data(
        ImageType.SEG2, FAKE_IMAGE_DATA
    )

    # Assert
    # since not all images have finished loading, expect call to next to raise error
    with pytest.raises(RuntimeError):
        curation_model_main_view.next_image()

    # Act
    curation_model_main_view.set_curr_image_data(
        ImageType.SEG2, FAKE_IMAGE_DATA
    )
    curation_model_main_view.next_image()

    # Assert
    assert (
        cursor_moved_slot.call_count == 2
    )  # once after start, once after next
    assert img_loading_finished_slot.call_count == 1  # once after start

    # Act
    curation_model_main_view.set_next_image_data(
        ImageType.RAW, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_next_image_data(
        ImageType.SEG1, FAKE_IMAGE_DATA
    )
    curation_model_main_view.set_next_image_data(
        ImageType.SEG2, FAKE_IMAGE_DATA
    )

    # Assert
    assert img_loading_finished_slot.call_count == 2

    # Act
    curation_model_main_view.next_image()
    # we are at the last image, so there is no next image data to load

    # Assert
    assert cursor_moved_slot.call_count == 3
    assert img_loading_finished_slot.call_count == 3
    # there is no next image
    with pytest.raises(RuntimeError):
        curation_model_main_view.next_image()


# TODO: if we end up needing to change back from main to input view, add tests for that
# here


def test_set_merging_mask(
    curation_model_loading_started: CurationModel,
) -> None:
    # Arrange
    mask: np.ndarray = np.asarray([[1, 2], [3, 4]])
    # Act
    curation_model_loading_started.set_merging_mask(mask)
    # Assert
    assert np.array_equal(
        curation_model_loading_started.get_merging_mask(), mask
    )


def test_set_excluding_mask(
    curation_model_loading_started: CurationModel,
) -> None:
    # Arrange
    mask: np.ndarray = np.asarray([[8, 2], [3, 5]])
    # Act
    curation_model_loading_started.set_excluding_mask(mask)
    # Assert
    assert np.array_equal(
        curation_model_loading_started.get_excluding_mask(), mask
    )


def test_set_base_image(curation_model_loading_started: CurationModel) -> None:
    # Act
    curation_model_loading_started.set_base_image("seg2")

    # Assert
    assert curation_model_loading_started.get_base_image() == "seg2"


def test_set_use_image(curation_model_loading_started: CurationModel) -> None:
    # Act
    curation_model_loading_started.set_use_image(False)

    # Assert
    assert not curation_model_loading_started.get_use_image()


def test_set_curation_saved_to_disk(
    curation_model_loading_started: CurationModel,
) -> None:
    # Arrange
    saved_slot: Mock = Mock()
    curation_model_loading_started.saved_to_disk.connect(saved_slot)

    # Act
    curation_model_loading_started.set_curation_record_saved_to_disk(False)
    # Assert
    saved_slot.assert_called_with(False)

    # Act
    curation_model_loading_started.set_curation_record_saved_to_disk(True)
    # Assert
    saved_slot.assert_called_with(True)


def test_state_resets_on_next(
    curation_model_loading_started: CurationModel,
) -> None:
    # Act
    curation_model_loading_started.set_base_image("seg2")
    curation_model_loading_started.set_merging_mask(
        np.asarray([[1, 2], [3, 4]])
    )
    curation_model_loading_started.set_excluding_mask(
        np.asarray([[2, 3], [4, 5]])
    )
    curation_model_loading_started.set_use_image(False)
    curation_model_loading_started.next_image()

    # Assert
    assert curation_model_loading_started.get_base_image() == "seg1"
    assert curation_model_loading_started.get_merging_mask() == None
    assert curation_model_loading_started.get_excluding_mask() == None
    assert curation_model_loading_started.get_use_image() == True


def test_save_curation_record_to_disk(
    curation_model_loading_started: CurationModel,
) -> None:
    # Arrange
    save_requested_slot: Mock = Mock()
    curation_model_loading_started.save_to_disk_requested.connect(
        save_requested_slot
    )

    # Act
    curation_model_loading_started.save_curr_curation_record_to_disk()

    # Assert
    save_requested_slot.assert_called_once()



import os
import platform
import subprocess
import random
from csv import DictReader
from pathlib import Path
from typing import List, Generator, Tuple, Optional

from allencell_ml_segmenter.curation.curation_data_class import CurationRecord
from allencell_ml_segmenter.utils.file_writer import IFileWriter
from allencell_ml_segmenter.main.main_model import MIN_DATASET_SIZE

LOSS_COLUMN: str = "val/loss_epoch"


class FileUtils:
    """
    FileUtils handles file reading/writing tasks. In order to use the instance methods (write methods),
    please initialize a FileUtils instance with an IFileWriter object.
    """

    def __init__(self, file_writer: IFileWriter):
        self._file_writer = file_writer

    @staticmethod
    def get_all_files_in_dir_ignore_hidden(dir_path: Path) -> List[Path]:
        # sort alphabetically- default sorting behavior for glob
        all_files: List[Path] = list(sorted(dir_path.glob("*.*")))
        # Ignore hidden files (such as .DS_Store on mac)
        # There's no way to do this with Path.glob filtering or methods so using list comprehension
        return [file for file in all_files if not file.name.startswith(".")]

    @staticmethod
    def get_img_path_from_folder(folder: Path) -> Path:
        """
        Returns path of an image in the folder.
        :param folder: path to a folder containing images
        """
        # we expect user to have the same number of channels for all images in their folders
        # and that only images are stored in those folders

        path_generator: Generator[Path] = folder.glob("*.*")
        image: Path = next(path_generator)
        # ignore hidden files
        while str(image.name).startswith("."):
            image = next(path_generator)
        return image.resolve()

    @staticmethod
    def count_images_in_csv_folder(folder: Path) -> int:
        """
        Given a :param folder: containing train/test/val csvs, returns the
        number of unique images contained in all of the csvs.
        """
        path_generator: Generator[Path] = folder.glob("*.csv")
        images: set[str] = set()
        for p in path_generator:
            with open(p, newline="") as fr:
                reader: DictReader = DictReader(fr)
                for row in reader:
                    if "raw" in row:
                        images.add(str(Path(row["raw"]).resolve()))
        return len(images)

    @staticmethod
    def get_min_loss_from_csv(csv_path: Path) -> Optional[float]:
        min_loss: Optional[float] = None
        with open(csv_path, newline="") as fr:
            reader: DictReader = DictReader(fr)
            if (
                reader.fieldnames is None
                or LOSS_COLUMN not in reader.fieldnames
            ):
                return None

            for row in reader:
                entry: str = row[LOSS_COLUMN]
                if len(entry) > 0:
                    loss: float = float(entry)
                    if min_loss is None or loss < min_loss:
                        min_loss = loss

        return min_loss

    def write_curation_record(
        self,
        curation_records: List[CurationRecord],
        csv_dir_path: Path,
        mask_dir_path: Path,
    ) -> None:
        """
        Saves the curation record as a train and test csv in csv_path_dir and associated masks in mask_dir_path
        :param curation_record: record to save to csv
        :param csv_dir_path: directory to save csv (csvs will be named train.csv and test.csv)
        :param mask_dir_path: directory in which to save masks (masks will be saved under excluding_masks or
        merging_masks subdirs)
        """
        train, test = self._train_test_split(curation_records)
        self._write_curation_csv(
            train, csv_dir_path / "train.csv", mask_dir_path
        )
        self._write_curation_csv(
            test, csv_dir_path / "test.csv", mask_dir_path
        )
        self._write_curation_csv(test, csv_dir_path / "val.csv", mask_dir_path)

    def _train_test_split(
        self,
        curation_records: List[CurationRecord],
    ) -> Tuple[List[CurationRecord], List[CurationRecord]]:
        """
        Returns (train_split, test_split) of the items marked to_use in the provided curation record.
        Attempts to reserve 10% of these for test, but if that value is < 2, it will become 2, and
        if it is > 100, it will become 100. If there are < 4 curation records selected for use, an exception will be thrown.
        :param curation_record: record to split
        """
        curation_records_to_use: List[CurationRecord] = [
            r for r in curation_records if r.to_use
        ]
        if len(curation_records_to_use) < MIN_DATASET_SIZE:
            raise RuntimeError(
                f"At least {MIN_DATASET_SIZE} images must be selected for use"
            )

        test_len: int = max(2, min(100, len(curation_records_to_use) // 10))
        random.shuffle(curation_records_to_use)
        return (
            curation_records_to_use[test_len:],
            curation_records_to_use[:test_len],
        )

    def _write_curation_csv(
        self,
        curation_records: List[CurationRecord],
        csv_path: Path,
        mask_dir_path: Path,
    ) -> None:
        """
        Saves the curation record as a csv at csv_path and associated masks under mask_dir_path
        :param curation_record: record to save to csv
        :param csv_path: path to save csv
        :param mask_dir_path: directory in which to save masks (masks will be saved under excluding_masks or
        merging_masks subdirs)
        """
        self._file_writer.csv_open_write_mode(csv_path)
        self._file_writer.csv_write_row(
            csv_path,
            [
                "",
                "raw",
                "seg1",
                "seg2",
                "merge_mask",
                "exclude_mask",
                "base_image",
            ],
        )

        get_excl_mask_path = (
            lambda raw_path: mask_dir_path
            / "excluding_masks"
            / f"excluding_mask_{raw_path.stem}.npy"
        )
        get_merg_mask_path = (
            lambda raw_path: mask_dir_path
            / "merging_masks"
            / f"merging_mask_{raw_path.stem}.npy"
        )

        idx = 0
        for record in curation_records:
            if record.to_use:
                if record.excluding_mask is not None:
                    self._file_writer.np_save(
                        get_excl_mask_path(record.raw_file.resolve()),
                        record.excluding_mask,
                    )
                if record.merging_mask is not None:
                    self._file_writer.np_save(
                        get_merg_mask_path(record.raw_file.resolve()),
                        record.merging_mask,
                    )

                self._file_writer.csv_write_row(
                    csv_path,
                    [
                        str(idx),
                        str(record.raw_file.resolve()),
                        str(record.seg1.resolve()),
                        (
                            str(record.seg2.resolve())
                            if record.seg2 is not None
                            else ""
                        ),
                        (
                            str(get_merg_mask_path(record.raw_file.resolve()))
                            if record.merging_mask is not None
                            else ""
                        ),
                        (
                            str(get_excl_mask_path(record.raw_file.resolve()))
                            if record.excluding_mask is not None
                            else ""
                        ),
                        str(record.base_image),
                    ],
                )
                idx += 1
        self._file_writer.csv_close(csv_path)

    @staticmethod
    def open_directory_in_window(dir: Path) -> None:
        # for Windows operating systems
        if platform.system() == "Windows":
            os.startfile(dir)  # type: ignore
        # for MacOS operating systems
        elif platform.system() == "Darwin":
            subprocess.Popen(["open", dir])
        # for Linux
        else:
            subprocess.Popen(["xdg-open", dir])



from qtpy.QtWidgets import QWidget

from allencell_ml_segmenter.core.subscriber import Subscriber


class AicsWidgetMeta(type(QWidget), type(Subscriber)):  # type: ignore
    pass


class AicsWidget(QWidget, Subscriber, metaclass=AicsWidgetMeta):
    """
    Base class for custom widgets to inherit from
    """

    _template = None

    def __init__(self) -> None:
        QWidget.__init__(self)



from unittest.mock import patch

import pytest
from qtpy.QtWidgets import QFileDialog
from pytestqt.qtbot import QtBot

from allencell_ml_segmenter.widgets.directory_or_csv_file_dialog import (
    DirectoryOrCSVFileDialog,
)


@pytest.fixture
def directory_or_csv_file_dialog(qtbot: QtBot) -> DirectoryOrCSVFileDialog:
    """
    Fixture that creates an instance of CustomFileDialog for testing.
    """
    return DirectoryOrCSVFileDialog()


def test_selected(
    directory_or_csv_file_dialog: DirectoryOrCSVFileDialog,
) -> None:
    """
    Test the _selected method of CustomFileDialog.
    """
    # ARRANGE
    mock_path: str = "/path/to/file"

    # ACT
    with patch("os.path.isdir", return_value=True):
        directory_or_csv_file_dialog._selected(mock_path)

    # ASSERT
    assert directory_or_csv_file_dialog.fileMode() == QFileDialog.Directory

    # ACT
    with patch("os.path.isdir", return_value=False):
        directory_or_csv_file_dialog._selected(mock_path)

    # ASSERT
    assert directory_or_csv_file_dialog.fileMode() == QFileDialog.ExistingFile



import pytest
from allencell_ml_segmenter.core.publisher import Publisher, Event
from allencell_ml_segmenter._tests.fakes.fake_subscriber import FakeSubscriber


@pytest.fixture
def publisher() -> Publisher:
    """
    Fixture for Publisher.
    """
    return Publisher()


def test_pub_dispatch(publisher: Publisher) -> None:
    """
    Tests that a publisher can properly dispatch an event to a subscriber.
    """
    # ARRANGE
    subscriber: FakeSubscriber = FakeSubscriber()
    event_under_test: Event = Event.PROCESS_TRAINING
    publisher.subscribe(event_under_test, subscriber, subscriber.handle)

    # ACT
    publisher.dispatch(event_under_test)

    # ASSERT
    assert subscriber.was_handled(event_under_test)


# TODO from brian: add better methods for testing publisher event to subscriber handlers


def test_pub_dispatch_multiple(publisher: Publisher) -> None:
    """
    Tests that a publisher can properly dispatch an event to multiple subscribers.
    """

    # ARRANGE
    subscriber1: FakeSubscriber = FakeSubscriber()
    subscriber2: FakeSubscriber = FakeSubscriber()

    publisher.subscribe(
        Event.PROCESS_TRAINING, subscriber1, subscriber1.handle
    )
    publisher.subscribe(
        Event.PROCESS_TRAINING, subscriber2, subscriber2.handle
    )

    # ACT
    publisher.dispatch(Event.PROCESS_TRAINING)

    # ASSERT
    assert subscriber1.was_handled(Event.PROCESS_TRAINING)
    assert subscriber2.was_handled(Event.PROCESS_TRAINING)


def test_pub_unsubscribe(publisher: Publisher) -> None:
    """
    Tests that a publisher can properly unsubscribe a subscriber.
    """

    # ARRANGE
    subscriber: FakeSubscriber = FakeSubscriber()
    event_under_test: Event = Event.PROCESS_TRAINING

    publisher.subscribe(event_under_test, subscriber, subscriber.handle)
    publisher.unsubscribe(event_under_test, subscriber)

    # ACT
    publisher.dispatch(event_under_test)

    # ASSERT
    assert subscriber.was_handled(event_under_test) is False


def test_pub_unsubscribe_unknown(publisher: Publisher) -> None:
    """
    Tests that unsubscribing a nonexistent subscriber does not cause errors.
    """
    subscriber: FakeSubscriber = FakeSubscriber()

    # ACT
    publisher.unsubscribe(Event.PROCESS_TRAINING, subscriber)

    # If no error is raised, the test will pass (no arbitrary assertion needed)



from allencell_ml_segmenter.core.publisher import Publisher
from allencell_ml_segmenter.core.event import Event
from enum import Enum
from typing import Union, Optional, List
from pathlib import Path
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.main.main_model import MainModel, ImageType
from qtpy.QtCore import QObject, Signal
from allencell_ml_segmenter.utils.experiment_utils import ExperimentUtils


class TrainingType(Enum):
    """
    Different cyto-dl experiment types
    """

    SEGMENTATION_PLUGIN = "segmentation_plugin"
    SEGMENTATION = "segmentation"
    GAN = "gan"
    OMNIPOSE = "omnipose"
    SKOOTS = "skoots"


class ModelSize(Enum):
    """
    Model size for training, and their respective filters overrides
    """

    SMALL = [8, 16, 32]
    MEDIUM = [16, 32, 64]
    LARGE = [32, 64, 128]


# TODO: move these signals directly into TrainingModel once we deprecate
# 'Publisher' in a refactor
class TrainingModelSignals(QObject):
    num_channels_set: Signal = Signal()
    images_directory_set: Signal = Signal()
    spatial_dims_set: Signal = Signal()


class TrainingModel(Publisher):
    """
    Stores state relevant to training processes.
    """

    def __init__(
        self, main_model: MainModel, experiments_model: ExperimentsModel
    ):
        super().__init__()
        self._main_model: MainModel = main_model
        self.experiments_model: ExperimentsModel = experiments_model
        self.signals: TrainingModelSignals = TrainingModelSignals()
        self._experiment_type: TrainingType = TrainingType.SEGMENTATION_PLUGIN
        self._images_directory: Optional[Path] = None
        self._selected_channel: dict[ImageType, Optional[int]] = {
            t: None for t in ImageType
        }
        self._num_channels: dict[ImageType, Optional[int]] = {
            t: None for t in ImageType
        }
        self._model_path: Optional[Path] = None  # if None, start a new model
        self._patch_size: Optional[list[int]] = None
        self._spatial_dims: Optional[int] = None
        self._num_epochs: Optional[int] = None
        self._max_time: Optional[int] = None  # in minutes
        self._max_channel: Optional[int] = None
        self._use_max_time: bool = (
            False  # default is false. UI starts with max epoch defined rather than max time.
        )
        self._model_size: Optional[ModelSize] = None
        # the total number of images used for training/test/validation for this model
        self._total_num_images: int = 0

        # Whether to use an existing model, and the existing model to use if one is selected
        # If is_using_existing_model is False, the existing_model_to_use will be None
        # If is_using_existing_model is True, the existing_model_to_use will be the name of the experiment containing
        # the existing model to pull weights from for training
        # If existing_model_to_use is None, no model was selected to pull weights from and the user should be prompted to select one
        self._is_using_existing_model: bool = False  # No by default
        self._existing_model_to_use: Optional[str] = (
            None  # None if no existing model selected- default behavior
        )

    def get_experiment_type(self) -> Optional[str]:
        """
        Gets experiment type
        """
        if self._experiment_type is None:
            return None
        return self._experiment_type.value

    def set_experiment_type(self, training_type: str) -> None:
        """
        Sets experiment type

        training_type (str): name of cyto-dl experiment to run
        """
        # convert string to enum
        self._experiment_type = TrainingType(training_type)

    def get_spatial_dims(self) -> Optional[int]:
        """
        Gets image dimensions
        """
        return self._spatial_dims

    def set_spatial_dims(self, spatial_dims: int) -> None:
        """
        Sets image dimensions

        image_dims (int): number of dimensions to train model on. "2" for 2D, "3" for 3D
        """
        if spatial_dims != 2 and spatial_dims != 3:
            raise ValueError("No support for non 2D and 3D images.")
        self._spatial_dims = spatial_dims
        self.signals.spatial_dims_set.emit()

    def get_num_epochs(self) -> Optional[int]:
        """
        Gets max epoch
        """
        return self._num_epochs

    def set_num_epochs(self, num_epochs: int) -> None:
        """
        Sets num epochs

        num_epochs (int): number of additional epochs to train for
        """
        self._num_epochs = num_epochs

    def get_images_directory(self) -> Optional[Path]:
        """
        Gets images directory
        """
        return self._images_directory

    def set_images_directory(self, images_path: Optional[Path]) -> None:
        """
        Sets images directory, and dispatches channel extraction

        images_path (Path): path to images directory
        """
        self._images_directory = images_path
        self.signals.images_directory_set.emit()

    def get_num_channels(self, image_type: ImageType) -> Optional[int]:
        if self._num_channels is not None:
            return self._num_channels[image_type]
        return None

    def set_all_num_channels(
        self, num_channels: dict[ImageType, Optional[int]]
    ) -> None:
        self._num_channels = num_channels
        self.signals.num_channels_set.emit()

    def get_selected_channel(self, image_type: ImageType) -> Optional[int]:
        return self._selected_channel[image_type]

    def set_selected_channel(
        self, image_type: ImageType, channel: Optional[int]
    ) -> None:
        self._selected_channel[image_type] = channel

    def get_patch_size(self) -> Optional[list[int]]:
        """
        Gets patch size
        """
        return self._patch_size

    def set_patch_size(self, patch_size: list[int]) -> None:
        """
        Sets patch size

        patch_size (str): patch size for training
        """
        if len(patch_size) not in [2, 3]:
            raise ValueError(
                "Patch sizes need to be 2 or 3 dimension based on input image."
            )
        self._patch_size = patch_size

    def get_max_time(self) -> Optional[int]:
        """
        Gets max runtime (in seconds)
        """
        return self._max_time

    def set_max_time(self, max_time: int) -> None:
        """
        Sets max runtime (in seconds)

        max_time (int): maximum runtime for training, in seconds
        """
        self._max_time = max_time

    def dispatch_training(self) -> None:
        """
        Dispatches even to start training
        """
        self.dispatch(Event.PROCESS_TRAINING)

    def use_max_time(self) -> bool:
        """
        Will training run will be based off of max time
        """
        return self._use_max_time

    def set_use_max_time(self, use_max: bool) -> None:
        """
        Set if training run will be based off of max time
        """
        self._use_max_time = use_max

    def set_model_size(self, model_size: Optional[str]) -> None:
        if model_size is None or model_size == "":
            self._model_size = None
        else:
            # convert string to enum
            model_size = model_size.upper()
            if model_size not in [x.name for x in ModelSize]:
                raise ValueError(
                    "No support for non small, medium, and large patch sizes."
                )
            self._model_size = ModelSize[model_size]

    def get_model_size(self) -> Optional[ModelSize]:
        return self._model_size

    def set_total_num_images(self, num_images: int) -> None:
        self._total_num_images = num_images

    def get_total_num_images(self) -> int:
        return self._total_num_images

    def get_selected_channels(self) -> dict[ImageType, Optional[int]]:
        return self._main_model.get_selected_channels()

    def set_existing_model(self, model: Optional[str]) -> None:
        """
        Set the existing model to use for iterative training
        """
        self._existing_model_to_use = model

    def get_existing_model(self) -> Optional[str]:
        """
        Existing model to be used for iterative training
        """
        return self._existing_model_to_use

    def is_using_existing_model(self) -> bool:
        """
        Iterative training- will use existing model to start
        """
        return self._is_using_existing_model

    def set_is_using_existing_model(self, is_using: bool) -> None:
        """
        Set iterative training
        """
        self._is_using_existing_model = is_using

    def get_existing_model_ckpt_path(self) -> Optional[Path]:
        model_to_use: Optional[str] = self.get_existing_model()
        user_exp_path: Optional[Path] = (
            self.experiments_model.get_user_experiments_path()
        )

        if model_to_use is None or user_exp_path is None:
            # shouldn't be possible to get here
            return None
        return ExperimentUtils.get_best_ckpt(user_exp_path, model_to_use)



from pathlib import Path

from allencell_ml_segmenter.core.channel_extraction import (
    get_img_path_from_csv,
)
from allencell_ml_segmenter.core.image_data_extractor import (
    IImageDataExtractor,
    AICSImageDataExtractor,
    ImageData,
)
from allencell_ml_segmenter.core.subscriber import Subscriber
from allencell_ml_segmenter.core.event import Event

from cyto_dl.api.model import CytoDLModel  # type: ignore
from allencell_ml_segmenter.main.experiments_model import ExperimentsModel
from allencell_ml_segmenter.training.training_model import (
    TrainingModel,
    ImageType,
)
from typing import Optional
from napari.utils.notifications import show_warning, show_error  # type: ignore
from allencell_ml_segmenter.utils.cyto_overrides_manager import (
    CytoDLOverridesManager,
)
from allencell_ml_segmenter.utils.file_utils import FileUtils
from allencell_ml_segmenter.core.task_executor import (
    ITaskExecutor,
    NapariThreadTaskExecutor,
)
from collections import namedtuple
from allencell_ml_segmenter.main.main_model import MIN_DATASET_SIZE


DirectoryData = namedtuple(
    "DirectoryData",
    [
        "num_images",
        "spatial_dims",
        "raw_channels",
        "seg1_channels",
        "seg2_channels",
    ],
)


class TrainingService(Subscriber):
    """
    Interface for training a model. Uses cyto-dl to train model according to spec
    """

    def __init__(
        self,
        training_model: TrainingModel,
        experiments_model: ExperimentsModel,
        img_data_extractor: IImageDataExtractor = AICSImageDataExtractor.global_instance(),
        task_executor: ITaskExecutor = NapariThreadTaskExecutor.global_instance(),
    ):
        super().__init__()
        self._training_model: TrainingModel = training_model
        self._experiments_model: ExperimentsModel = experiments_model
        self._task_executor: ITaskExecutor = task_executor
        self._training_model.subscribe(
            Event.PROCESS_TRAINING,
            self,
            self._train_model_handler,
        )
        self._img_data_extractor: IImageDataExtractor = img_data_extractor
        self._training_model.signals.images_directory_set.connect(
            self._training_image_directory_selected
        )

    def _train_model_handler(self, _: Event) -> None:
        """
        Trains the model according to the spec
        """
        # Only supporting segmentation config for now, in the future this will be an option in the UI
        self._training_model.set_experiment_type("segmentation_plugin")
        # TODO make set_images_directory and get_images_directory less brittle.
        #  https://github.com/AllenCell/allencell-ml-segmenter/issues/156
        if self._able_to_continue_training():
            cyto_dl_model = CytoDLModel()
            if self._training_model.is_using_existing_model():
                # ITERATIVE TRAINING: train starting from existing model weights
                cyto_dl_model.load_config_from_file(
                    str(
                        self._experiments_model.get_train_config_path(
                            self._training_model.get_existing_model()
                        )
                    )
                )
            else:
                # NEW TRAINING: load the default experiment config
                cyto_dl_model.load_default_experiment(
                    self._training_model.get_experiment_type(),
                    output_dir=f"{self._experiments_model.get_user_experiments_path()}/{self._experiments_model.get_experiment_name()}",
                )
            cyto_overrides_manager: CytoDLOverridesManager = (
                CytoDLOverridesManager(
                    self._experiments_model, self._training_model
                )
            )
            cyto_dl_model.override_config(
                cyto_overrides_manager.get_training_overrides()
            )
            cyto_dl_model.print_config()
            cyto_dl_model.save_config(
                self._experiments_model.get_train_config_path()
            )
            cyto_dl_model.train()

    def _able_to_continue_training(self) -> bool:
        # TODO: refactor- these checks should be in the View before we start a thread for training.
        if self._experiments_model.get_experiment_name() is None:
            show_warning(
                "Please select an experiment before running prediction."
            )
            return False

        if self._training_model.get_experiment_type() is None:
            return False

        if self._training_model.get_spatial_dims() is None:
            show_warning(
                "Please select spatial dims for training dataset. 2-D or 3-D."
            )
            return False

        if self._training_model.get_images_directory() is None:
            show_warning("User has not selected input images for training")
            return False

        if self._training_model.get_patch_size() is None:
            show_warning("User has not selected a patch size for training")
            return False

        if self._training_model.get_num_epochs() is None:
            show_warning("Please define max epoch(s) to run for")
            return False

        if (
            self._training_model.get_model_size() is None
            and not self._training_model.is_using_existing_model()
        ):
            show_warning(
                "Please define model size for a new model from scratch."
            )
            return False

        if self._training_model.is_using_existing_model():
            if self._training_model.get_existing_model() is None:
                show_warning(
                    "If using weights from an existing model, please select one."
                )
                return False
        return True

    def _extract_data_from_training_dir(
        self, training_dir: Path
    ) -> DirectoryData:
        num_imgs: int = FileUtils.count_images_in_csv_folder(training_dir)
        if num_imgs < MIN_DATASET_SIZE:
            raise RuntimeError(
                f"Training requires at least {MIN_DATASET_SIZE} images and their segmentations"
            )

        training_csv: Path = training_dir / "train.csv"
        raw_data: ImageData = self._img_data_extractor.extract_image_data(
            get_img_path_from_csv(training_csv, column="raw"), np_data=False
        )
        seg1_data: ImageData = self._img_data_extractor.extract_image_data(
            get_img_path_from_csv(training_csv, column="seg1"), np_data=False
        )
        seg2_data: Optional[ImageData]
        try:
            seg2_path: Path = get_img_path_from_csv(training_csv, "seg2")
            seg2_data = self._img_data_extractor.extract_image_data(
                seg2_path, np_data=False
            )
        except ValueError:
            seg2_data = None

        return DirectoryData(
            num_imgs,
            3 if raw_data.dim_z is not None and raw_data.dim_z > 1 else 2,
            raw_data.channels,
            seg1_data.channels,
            seg2_data.channels if seg2_data else None,
        )

    def _on_training_dir_data_extracted(self, dir_data: DirectoryData) -> None:
        self._training_model.set_total_num_images(dir_data.num_images)
        self._training_model.set_all_num_channels(
            {
                ImageType.RAW: dir_data.raw_channels,
                ImageType.SEG1: dir_data.seg1_channels,
                ImageType.SEG2: dir_data.seg2_channels,
            }
        )
        self._training_model.set_spatial_dims(dir_data.spatial_dims)

    def _on_training_dir_data_error(self, e: Exception) -> None:
        self._training_model.set_total_num_images(0)
        self._training_model.set_all_num_channels(
            {
                ImageType.RAW: None,
                ImageType.SEG1: None,
                ImageType.SEG2: None,
            }
        )
        show_error(f"Failed to get data from training directory: {e}")

    def _training_image_directory_selected(self) -> None:
        training_dir: Optional[Path] = (
            self._training_model.get_images_directory()
        )
        if training_dir is not None:
            self._task_executor.exec(
                lambda: self._extract_data_from_training_dir(training_dir),
                on_return=self._on_training_dir_data_extracted,
                on_error=self._on_training_dir_data_error,
            )
